"""Copyright 2019-2026, XGBoost contributors"""

import asyncio
import json
import os
import pickle
import socket
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Tuple, Type, Union

import dask
import dask.array as da
import dask.dataframe as dd
import distributed
import hypothesis
import numpy as np
import pytest
import scipy
import sklearn
import xgboost as xgb
from distributed import Client, LocalCluster, Nanny, Worker
from distributed.scheduler import KilledWorker, Scheduler
from distributed.utils_test import async_poll_for, gen_cluster
from hypothesis import HealthCheck, assume, given, note, settings
from sklearn.datasets import make_classification, make_regression
from sklearn.model_selection import train_test_split
from xgboost import collective as coll
from xgboost import dask as dxgb
from xgboost import testing as tm
from xgboost.collective import Config as CollConfig
from xgboost.dask import DaskDMatrix
from xgboost.testing.dask import (
    check_init_estimation,
    check_uneven_nan,
    get_rabit_args,
    make_categorical,
    run_recode,
)
from xgboost.testing.data import get_california_housing
from xgboost.testing.params import hist_cache_strategy, hist_parameter_strategy
from xgboost.testing.shared import (
    get_feature_weights,
    validate_data_initialization,
    validate_leaf_output,
)
from xgboost.testing.updater import get_basescore

dask.config.set({"distributed.scheduler.allowed-failures": False})

pytestmark = tm.timeout(60)


if hasattr(HealthCheck, "function_scoped_fixture"):
    suppress = [HealthCheck.function_scoped_fixture]
else:
    suppress = hypothesis.utils.conventions.not_set  # type: ignore


kRows = 1000
kCols = 10
kWorkers = 5


def generate_array(
    with_weights: bool = False,
) -> Tuple[da.Array, da.Array, Optional[da.Array]]:
    chunk_size = 20
    rng = da.random.RandomState(1994)
    X = rng.random_sample((kRows, kCols), chunks=(chunk_size, -1))
    y = rng.random_sample(kRows, chunks=chunk_size)
    if with_weights:
        w = rng.random_sample(kRows, chunks=chunk_size)
        return X, y, w
    return X, y, None


@pytest.mark.parametrize("to_frame", [True, False])
def test_xgbclassifier_classes_type_and_value(to_frame: bool, client: "Client") -> None:
    X, y = make_classification(n_samples=1000, n_features=4, random_state=123)
    if to_frame:
        import pandas as pd

        feats = [f"var_{i}" for i in range(4)]
        df = pd.DataFrame(X, columns=feats)
        df["target"] = y
        df = dd.from_pandas(df, npartitions=1)
        X, y = df[feats], df["target"]
    else:
        X = da.from_array(X)
        y = da.from_array(y)

    est = dxgb.DaskXGBClassifier(n_estimators=10).fit(X, y)
    assert isinstance(est.classes_, np.ndarray)
    np.testing.assert_array_equal(est.classes_, np.array([0, 1]))


def test_from_dask_dataframe(client: "Client") -> None:
    X_, y_, _ = generate_array()

    X = dd.from_dask_array(X_)
    y = dd.from_dask_array(y_)

    dtrain = DaskDMatrix(client, X, y)
    booster = dxgb.train(client, {}, dtrain, num_boost_round=2)["booster"]

    prediction = dxgb.predict(client, model=booster, data=dtrain)

    assert prediction.ndim == 1
    assert isinstance(prediction, da.Array)
    assert prediction.shape[0] == kRows

    with pytest.raises(TypeError):
        # evals_result is not supported in dask interface.
        dxgb.train(  # type: ignore
            client, {}, dtrain, num_boost_round=2, evals_result={}
        )
    # force prediction to be computed
    from_dmatrix = prediction.compute()

    prediction = dxgb.predict(client, model=booster, data=X)
    from_df = prediction.compute()

    assert isinstance(prediction, dd.Series)
    assert np.all(prediction.compute().values == from_dmatrix)
    assert np.all(from_dmatrix == from_df.to_numpy())

    series_predictions = dxgb.inplace_predict(client, booster, X)
    assert isinstance(series_predictions, dd.Series)
    np.testing.assert_allclose(series_predictions.compute().values, from_dmatrix)

    # Make sure the output can be integrated back to original dataframe
    X["predict"] = prediction
    X["inplace_predict"] = series_predictions

    assert bool(X.isnull().values.any().compute()) is False


def test_from_dask_array(client: "Client") -> None:
    X, y, _ = generate_array()
    dtrain = DaskDMatrix(client, X, y)
    # results is {'booster': Booster, 'history': {...}}
    result = dxgb.train(client, {}, dtrain)

    prediction = dxgb.predict(client, result, dtrain)
    assert prediction.shape[0] == kRows

    assert isinstance(prediction, da.Array)
    # force prediction to be computed
    prediction = prediction.compute()

    booster: xgb.Booster = result["booster"]
    single_node_predt = booster.predict(xgb.DMatrix(X.compute()))
    np.testing.assert_allclose(prediction, single_node_predt)

    config = json.loads(booster.save_config())
    scheduler_info = client.scheduler_info()
    worker_nthreads = next(iter(scheduler_info["workers"].values()))["nthreads"]
    assert int(config["learner"]["generic_param"]["nthread"]) == worker_nthreads

    from_arr = dxgb.predict(client, model=booster, data=X)

    assert isinstance(from_arr, da.Array)
    assert np.all(single_node_predt == from_arr.compute())


def test_dask_sparse(client: "Client") -> None:
    X_, y_ = make_classification(n_samples=1000, n_informative=5, n_classes=3)
    rng = np.random.default_rng(seed=0)
    idx = rng.integers(low=0, high=X_.shape[0], size=X_.shape[0] // 4)
    X_[idx, :] = np.nan

    # numpy
    X, y = da.from_array(X_), da.from_array(y_)
    clf = dxgb.DaskXGBClassifier(tree_method="hist", n_estimators=10)
    clf.client = client
    clf.fit(X, y, eval_set=[(X, y)])
    dense_results = clf.evals_result()

    # scipy sparse
    X, y = da.from_array(X_).map_blocks(scipy.sparse.csr_matrix), da.from_array(y_)
    clf = dxgb.DaskXGBClassifier(tree_method="hist", n_estimators=10)
    clf.client = client
    clf.fit(X, y, eval_set=[(X, y)])
    sparse_results = clf.evals_result()
    np.testing.assert_allclose(
        dense_results["validation_0"]["mlogloss"],
        sparse_results["validation_0"]["mlogloss"],
    )


def run_categorical(
    client: "Client",
    tree_method: str,
    device: str,
    X: dd.DataFrame,
    X_onehot: dd.DataFrame,
    y: dd.Series,
    tmp_path: Path,
) -> None:
    # Force onehot
    parameters = {
        "tree_method": tree_method,
        "device": device,
        "max_cat_to_onehot": 9999,
    }
    rounds = 10
    m = dxgb.DaskDMatrix(client, X_onehot, y)
    by_etl_results = dxgb.train(
        client,
        parameters,
        m,
        num_boost_round=rounds,
        evals=[(m, "Train")],
    )["history"]

    m = dxgb.DaskDMatrix(client, X, y)
    output = dxgb.train(
        client,
        parameters,
        m,
        num_boost_round=rounds,
        evals=[(m, "Train")],
    )
    by_builtin_results = output["history"]

    np.testing.assert_allclose(
        np.array(by_etl_results["Train"]["rmse"]),
        np.array(by_builtin_results["Train"]["rmse"]),
        rtol=1e-3,
    )
    assert tm.non_increasing(by_builtin_results["Train"]["rmse"])

    check_model_counter = [0]

    def check_model_output(model: dxgb.Booster) -> None:
        path = tmp_path / f"model_{check_model_counter[0]}.json"
        check_model_counter[0] += 1
        model.save_model(path)
        with open(path, "r") as fd:
            categorical = json.load(fd)

        categories_sizes = np.array(
            categorical["learner"]["gradient_booster"]["model"]["trees"][-1][
                "categories_sizes"
            ]
        )
        assert categories_sizes.shape[0] != 0
        np.testing.assert_allclose(categories_sizes, 1)

    check_model_output(output["booster"])
    reg = dxgb.DaskXGBRegressor(
        n_estimators=10,
        tree_method=tree_method,
        device=device,
        # force onehot
        max_cat_to_onehot=9999,
    )
    reg.fit(X, y)

    check_model_output(reg.get_booster())

    reg = dxgb.DaskXGBRegressor(n_estimators=10, tree_method="exact")
    with pytest.raises(ValueError, match="distributed training"):
        reg.fit(X, y)
    # check partition based
    reg = dxgb.DaskXGBRegressor(
        n_estimators=10,
        tree_method=tree_method,
        device=device,
    )
    reg.fit(X, y, eval_set=[(X, y)])
    assert tm.non_increasing(reg.evals_result()["validation_0"]["rmse"])

    booster = reg.get_booster()
    predt = dxgb.predict(client, booster, X).compute().values
    inpredt = dxgb.inplace_predict(client, booster, X).compute().values

    if hasattr(predt, "get"):
        predt = predt.get()
    if hasattr(inpredt, "get"):
        inpredt = inpredt.get()

    np.testing.assert_allclose(predt, inpredt)


def test_categorical(client: "Client", tmp_path: Path) -> None:
    X, y = make_categorical(client, 3000, 30, 13)
    X_onehot, _ = make_categorical(client, 3000, 30, 13, onehot=True)
    run_categorical(client, "approx", "cpu", X, X_onehot, y, tmp_path)
    run_categorical(client, "hist", "cpu", X, X_onehot, y, tmp_path)

    ft = ["c"] * X.shape[1]
    reg = dxgb.DaskXGBRegressor(tree_method="hist", feature_types=ft)
    reg.fit(X, y)
    assert reg.get_booster().feature_types == ft


def test_recode(client: "Client") -> None:
    run_recode(client, "cpu")


def test_dask_predict_shape_infer(client: "Client") -> None:
    X, y = make_classification(n_samples=kRows, n_informative=5, n_classes=3)
    X_ = dd.from_array(X, chunksize=100)
    y_ = dd.from_array(y, chunksize=100)
    dtrain = dxgb.DaskDMatrix(client, data=X_, label=y_)

    model = dxgb.train(
        client, {"objective": "multi:softprob", "num_class": 3}, dtrain=dtrain
    )

    preds = dxgb.predict(client, model, dtrain)
    assert preds.shape[0] == preds.compute().shape[0]
    assert preds.shape[1] == preds.compute().shape[1]

    prediction = dxgb.predict(client, model, X_, output_margin=True)
    assert isinstance(prediction, dd.DataFrame)

    prediction = prediction.compute()
    assert prediction.ndim == 2
    assert prediction.shape[0] == kRows
    assert prediction.shape[1] == 3

    prediction = dxgb.inplace_predict(client, model, X_, predict_type="margin")
    assert isinstance(prediction, dd.DataFrame)
    prediction = prediction.compute()
    assert prediction.ndim == 2
    assert prediction.shape[0] == kRows
    assert prediction.shape[1] == 3


def run_boost_from_prediction_multi_class(
    X: dd.DataFrame,
    y: dd.Series,
    tree_method: str,
    device: str,
    client: "Client",
) -> None:
    model_0 = dxgb.DaskXGBClassifier(
        learning_rate=0.3,
        n_estimators=4,
        tree_method=tree_method,
        max_bin=768,
        device=device,
    )
    model_0.fit(X=X, y=y, eval_set=[(X, y)])
    margin = dxgb.inplace_predict(
        client, model_0.get_booster(), X, predict_type="margin"
    )
    margin.columns = [f"m_{i}" for i in range(margin.shape[1])]

    model_1 = dxgb.DaskXGBClassifier(
        learning_rate=0.3,
        n_estimators=4,
        tree_method=tree_method,
        max_bin=768,
        device=device,
    )
    model_1.fit(
        X=X, y=y, base_margin=margin, eval_set=[(X, y)], base_margin_eval_set=[margin]
    )
    predictions_1 = dxgb.predict(
        client,
        model_1.get_booster(),
        dxgb.DaskDMatrix(client, X, base_margin=margin),
        output_margin=True,
    )

    model_2 = dxgb.DaskXGBClassifier(
        learning_rate=0.3,
        n_estimators=8,
        tree_method=tree_method,
        max_bin=768,
        device=device,
    )
    model_2.fit(X=X, y=y, eval_set=[(X, y)])
    predictions_2 = dxgb.inplace_predict(
        client, model_2.get_booster(), X, predict_type="margin"
    )
    a = predictions_1.compute()
    b = predictions_2.compute()
    # cupy/cudf
    if hasattr(a, "get"):
        a = a.get()
    if hasattr(b, "values"):
        b = b.values
    if hasattr(b, "get"):
        b = b.get()
    np.testing.assert_allclose(a, b, atol=1e-5)


def run_boost_from_prediction(
    X: dd.DataFrame,
    y: dd.Series,
    tree_method: str,
    device: str,
    client: "Client",
) -> None:
    X, y = client.persist([X, y])

    model_0 = dxgb.DaskXGBClassifier(
        learning_rate=0.3,
        n_estimators=3,
        tree_method=tree_method,
        max_bin=512,
        device=device,
    )
    model_0.fit(X=X, y=y, eval_set=[(X, y)])
    margin: dd.Series = model_0.predict(X, output_margin=True)

    model_1 = dxgb.DaskXGBClassifier(
        learning_rate=0.3,
        n_estimators=3,
        tree_method=tree_method,
        max_bin=512,
        device=device,
    )
    model_1.fit(
        X=X, y=y, base_margin=margin, eval_set=[(X, y)], base_margin_eval_set=[margin]
    )
    predictions_1: dd.Series = model_1.predict(X, base_margin=margin)

    model_2 = dxgb.DaskXGBClassifier(
        learning_rate=0.3,
        n_estimators=6,
        tree_method=tree_method,
        max_bin=512,
        device=device,
    )
    model_2.fit(X=X, y=y, eval_set=[(X, y)])
    predictions_2: dd.Series = model_2.predict(X)

    logloss_concat = (
        model_0.evals_result()["validation_0"]["logloss"]
        + model_1.evals_result()["validation_0"]["logloss"]
    )
    logloss_2 = model_2.evals_result()["validation_0"]["logloss"]
    np.testing.assert_allclose(logloss_concat, logloss_2, rtol=1e-4)

    margined = dxgb.DaskXGBClassifier(n_estimators=4)
    margined.fit(
        X=X, y=y, base_margin=margin, eval_set=[(X, y)], base_margin_eval_set=[margin]
    )

    unmargined = dxgb.DaskXGBClassifier(n_estimators=4)
    unmargined.fit(X=X, y=y, eval_set=[(X, y)], base_margin=margin)

    margined_res = margined.evals_result()["validation_0"]["logloss"]
    unmargined_res = unmargined.evals_result()["validation_0"]["logloss"]

    assert len(margined_res) == len(unmargined_res)
    for i in range(len(margined_res)):
        # margined is correct one, so smaller error.
        assert margined_res[i] < unmargined_res[i]


@pytest.mark.parametrize("tree_method", ["hist", "approx"])
def test_boost_from_prediction(tree_method: str, client_one_worker: "Client") -> None:
    from sklearn.datasets import load_breast_cancer, load_digits

    # This test has strict reproducibility requirements. However, Dask is freed to move
    # partitions between workers and modify the partitions' size during the test. Given
    # the lack of control over the partitioning logic, here we use a single worker as a
    # workaround.
    X_, y_ = load_breast_cancer(return_X_y=True)
    X, y = dd.from_array(X_, chunksize=200), dd.from_array(y_, chunksize=200)
    run_boost_from_prediction(X, y, tree_method, "cpu", client_one_worker)

    X_, y_ = load_digits(return_X_y=True)
    X_, _, y_, _ = train_test_split(
        X_, y_, train_size=300, stratify=y_, random_state=1994
    )
    X, y = dd.from_array(X_, chunksize=100), dd.from_array(y_, chunksize=100)
    run_boost_from_prediction_multi_class(X, y, tree_method, "cpu", client_one_worker)


def test_inplace_predict(client: "Client") -> None:
    from sklearn.datasets import load_diabetes

    X_, y_ = load_diabetes(return_X_y=True)
    X, y = dd.from_array(X_, chunksize=32), dd.from_array(y_, chunksize=32)
    reg = dxgb.DaskXGBRegressor(n_estimators=4)
    reg.client = client
    reg.fit(X, y)
    booster = reg.get_booster()
    base_margin = y

    inplace = client.compute(
        dxgb.inplace_predict(client, booster, X, base_margin=base_margin)
    ).result()
    Xy = dxgb.DaskDMatrix(client, X, base_margin=base_margin)
    copied = client.compute(dxgb.predict(client, booster, Xy)).result()
    np.testing.assert_allclose(inplace, copied)


def test_dask_missing_value_reg(client: "Client") -> None:
    X_0 = np.ones((20 // 2, kCols))
    X_1 = np.zeros((20 // 2, kCols))
    X = np.concatenate([X_0, X_1], axis=0)
    np.random.shuffle(X)
    X = da.from_array(X)
    X = X.rechunk(20, 1)
    y = da.random.randint(0, 3, size=20)
    y.rechunk(20)
    regressor = dxgb.DaskXGBRegressor(verbosity=1, n_estimators=2, missing=0.0)
    regressor.client = client
    regressor.set_params(tree_method="hist")
    regressor.fit(X, y, eval_set=[(X, y)])
    dd_predt = client.compute(regressor.predict(X)).result()

    np_X = client.compute(X).result()
    np_predt = regressor.get_booster().predict(xgb.DMatrix(np_X, missing=0.0))
    np.testing.assert_allclose(np_predt, dd_predt)


def test_dask_missing_value_cls(client: "Client") -> None:
    X_0 = np.ones((kRows // 2, kCols))
    X_1 = np.zeros((kRows // 2, kCols))
    X = np.concatenate([X_0, X_1], axis=0)
    np.random.shuffle(X)
    X = da.from_array(X)
    X = X.rechunk(20, None)
    y = da.random.randint(0, 3, size=kRows)
    y = y.rechunk(20, 1)
    cls = dxgb.DaskXGBClassifier(
        verbosity=1, n_estimators=2, tree_method="hist", missing=0.0
    )
    cls.client = client
    cls.fit(X, y, eval_set=[(X, y)])
    dd_pred_proba = cls.predict_proba(X).compute()

    np_X = X.compute()
    np_pred_proba = cls.get_booster().predict(xgb.DMatrix(np_X, missing=0.0))
    np.testing.assert_allclose(np_pred_proba, dd_pred_proba)

    cls = dxgb.DaskXGBClassifier()
    assert hasattr(cls, "missing")


@pytest.mark.parametrize("model", ["boosting", "rf"])
def test_dask_regressor(model: str, client: "Client") -> None:
    X, y, w = generate_array(with_weights=True)
    if model == "boosting":
        regressor = dxgb.DaskXGBRegressor(verbosity=1, n_estimators=2)
    else:
        regressor = dxgb.DaskXGBRFRegressor(verbosity=1, n_estimators=2)

    if hasattr(regressor, "_estimator_type"):
        assert regressor._estimator_type == "regressor"
    else:
        assert regressor._get_type() == "regressor"
    assert sklearn.base.is_regressor(regressor)

    regressor.set_params(tree_method="hist")
    regressor.client = client
    regressor.fit(X, y, sample_weight=w, eval_set=[(X, y)])
    prediction = regressor.predict(X)

    assert prediction.ndim == 1
    assert prediction.shape[0] == kRows

    history = regressor.evals_result()

    assert isinstance(prediction, da.Array)
    assert isinstance(history, dict)

    assert list(history["validation_0"].keys())[0] == "rmse"
    forest = int(
        json.loads(regressor.get_booster().save_config())["learner"][
            "gradient_booster"
        ]["gbtree_model_param"]["num_parallel_tree"]
    )

    if model == "boosting":
        assert len(history["validation_0"]["rmse"]) == 2
        assert forest == 1
    else:
        assert len(history["validation_0"]["rmse"]) == 1
        assert forest == 2


def run_dask_classifier(
    X: dxgb._DataT,
    y: dxgb._DaskCollection,
    w: dxgb._DaskCollection,
    model: str,
    tree_method: Optional[str],
    device: Literal["cpu", "cuda"],
    client: "Client",
    n_classes: int,
) -> None:
    metric = "merror" if n_classes > 2 else "logloss"

    if model == "boosting":
        classifier = dxgb.DaskXGBClassifier(
            verbosity=1,
            n_estimators=2,
            eval_metric=metric,
            tree_method=tree_method,
            device=device,
        )
    else:
        classifier = dxgb.DaskXGBRFClassifier(
            verbosity=1,
            n_estimators=2,
            eval_metric=metric,
            tree_method=tree_method,
            device=device,
        )

    if hasattr(classifier, "_estimator_type"):
        assert classifier._estimator_type == "classifier"
    else:
        assert classifier._get_type() == "classifier"
    assert sklearn.base.is_classifier(classifier)

    classifier.client = client
    classifier.fit(X, y, sample_weight=w, eval_set=[(X, y)])
    prediction = classifier.predict(X).compute()

    assert prediction.ndim == 1
    assert prediction.shape[0] == kRows

    history = classifier.evals_result()

    assert isinstance(history, dict)

    assert list(history.keys())[0] == "validation_0"
    assert list(history["validation_0"].keys())[0] == metric
    assert len(list(history["validation_0"])) == 1

    config = json.loads(classifier.get_booster().save_config())
    n_threads = int(config["learner"]["generic_param"]["nthread"])
    assert n_threads != 0 and n_threads != os.cpu_count()

    forest = int(
        config["learner"]["gradient_booster"]["gbtree_model_param"]["num_parallel_tree"]
    )
    if model == "boosting":
        assert len(history["validation_0"][metric]) == 2
        assert forest == 1
    else:
        assert len(history["validation_0"][metric]) == 1
        assert forest == 2

    # Test .predict_proba()
    probas = classifier.predict_proba(X).compute()
    assert classifier.n_classes_ == n_classes
    assert probas.ndim == 2
    assert probas.shape[0] == kRows
    assert probas.shape[1] == n_classes

    if n_classes > 2:
        cls_booster = classifier.get_booster()
        single_node_proba = cls_booster.inplace_predict(X.compute())

        # test shared by CPU and GPU
        if isinstance(single_node_proba, np.ndarray):
            np.testing.assert_allclose(single_node_proba, probas)
        else:
            import cupy

            cupy.testing.assert_allclose(single_node_proba, probas)

    # Test with dataframe, not shared with GPU as cupy doesn't work well with da.unique.
    if isinstance(X, da.Array) and n_classes > 2:
        X_d: dd.DataFrame = X.to_dask_dataframe()

        assert classifier.n_classes_ == n_classes
        prediction_df = classifier.predict(X_d).compute()

        assert prediction_df.ndim == 1
        assert prediction_df.shape[0] == kRows
        np.testing.assert_allclose(prediction_df, prediction)

        probas = classifier.predict_proba(X).compute()
        np.testing.assert_allclose(single_node_proba, probas)


@pytest.mark.parametrize("model", ["boosting", "rf"])
def test_dask_classifier(model: str, client: "Client") -> None:
    X, y, w = generate_array(with_weights=True)
    y = (y * 10).astype(np.int32)
    assert w is not None
    run_dask_classifier(X, y, w, model, None, "cpu", client, 10)

    y_bin = y.copy()
    y_bin[y > 5] = 1.0
    y_bin[y <= 5] = 0.0
    run_dask_classifier(X, y_bin, w, model, None, "cpu", client, 2)


def test_empty_dmatrix_training_continuation(client: "Client") -> None:
    kRows, kCols = 1, 97
    X = dd.from_array(np.random.randn(kRows, kCols))
    y = dd.from_array(np.random.rand(kRows))
    X.columns = ["X" + str(i) for i in range(0, kCols)]
    dtrain = dxgb.DaskDMatrix(client, X, y)

    kRows += 1000
    X = dd.from_array(np.random.randn(kRows, kCols), chunksize=10)
    X.columns = ["X" + str(i) for i in range(0, kCols)]
    y = dd.from_array(np.random.rand(kRows), chunksize=10)
    valid = dxgb.DaskDMatrix(client, X, y)

    out = dxgb.train(
        client,
        {"tree_method": "hist"},
        dtrain=dtrain,
        num_boost_round=2,
        evals=[(valid, "validation")],
    )

    out = dxgb.train(
        client,
        {"tree_method": "hist"},
        dtrain=dtrain,
        xgb_model=out["booster"],
        num_boost_round=2,
        evals=[(valid, "validation")],
    )
    assert dxgb.predict(client, out, dtrain).compute().shape[0] == 1


def run_empty_dmatrix_reg(client: "Client", parameters: dict) -> None:
    def _check_outputs(out: dxgb.TrainReturnT, predictions: np.ndarray) -> None:
        assert isinstance(out["booster"], dxgb.Booster)
        for _, v in out["history"]["validation"].items():
            assert len(v) == 2
        assert isinstance(predictions, np.ndarray)
        assert predictions.shape[0] == 1

    kRows, kCols = 1, 97
    X = dd.from_array(np.random.randn(kRows, kCols))
    y = dd.from_array(np.random.rand(kRows))
    dtrain = dxgb.DaskDMatrix(client, X, y)

    out = dxgb.train(
        client,
        parameters,
        dtrain=dtrain,
        evals=[(dtrain, "validation")],
        num_boost_round=2,
    )
    predictions = dxgb.predict(client=client, model=out, data=dtrain).compute()
    _check_outputs(out, predictions)

    # valid has more rows than train
    kRows += 1
    X = dd.from_array(np.random.randn(kRows, kCols))
    y = dd.from_array(np.random.rand(kRows))
    valid = dxgb.DaskDMatrix(client, X, y)
    out = dxgb.train(
        client,
        parameters,
        dtrain=dtrain,
        evals=[(valid, "validation")],
        num_boost_round=2,
    )
    predictions = dxgb.predict(client=client, model=out, data=dtrain).compute()
    _check_outputs(out, predictions)

    # train has more rows than evals
    valid = dtrain
    kRows += 1
    X = dd.from_array(np.random.randn(kRows, kCols))
    y = dd.from_array(np.random.rand(kRows))
    dtrain = dxgb.DaskDMatrix(client, X, y)

    out = dxgb.train(
        client,
        parameters,
        dtrain=dtrain,
        evals=[(valid, "validation")],
        num_boost_round=2,
    )
    predictions = dxgb.predict(client=client, model=out, data=valid).compute()
    _check_outputs(out, predictions)


def run_empty_dmatrix_cls(client: "Client", parameters: dict) -> None:
    n_classes = 4

    def _check_outputs(out: dxgb.TrainReturnT, predictions: np.ndarray) -> None:
        assert isinstance(out["booster"], dxgb.Booster)
        assert len(out["history"]["validation"]["merror"]) == 2
        assert isinstance(predictions, np.ndarray)
        assert predictions.shape[1] == n_classes, predictions.shape

    kRows, kCols = 1, 97
    X = dd.from_array(np.random.randn(kRows, kCols))
    y = dd.from_array(np.random.randint(low=0, high=n_classes, size=kRows))
    dtrain = dxgb.DaskDMatrix(client, X, y)
    parameters["objective"] = "multi:softprob"
    parameters["eval_metric"] = "merror"
    parameters["num_class"] = n_classes

    out = dxgb.train(
        client,
        parameters,
        dtrain=dtrain,
        evals=[(dtrain, "validation")],
        num_boost_round=2,
    )
    predictions = dxgb.predict(client=client, model=out, data=dtrain)
    assert predictions.shape[1] == n_classes
    predictions = predictions.compute()
    _check_outputs(out, predictions)

    # train has more rows than evals
    valid = dtrain
    kRows += 1
    X = dd.from_array(np.random.randn(kRows, kCols))
    y = dd.from_array(np.random.randint(low=0, high=n_classes, size=kRows))
    dtrain = dxgb.DaskDMatrix(client, X, y)

    out = dxgb.train(
        client,
        parameters,
        dtrain=dtrain,
        evals=[(valid, "validation")],
        num_boost_round=2,
    )
    predictions = dxgb.predict(client=client, model=out, data=valid).compute()
    _check_outputs(out, predictions)


def run_empty_dmatrix_auc(client: "Client", device: str, n_workers: int) -> None:
    from sklearn import datasets

    n_samples = 100
    n_features = 7
    rng = np.random.RandomState(1994)

    make_classification = partial(
        datasets.make_classification, n_features=n_features, random_state=rng
    )

    # binary
    X_, y_ = make_classification(n_samples=n_samples, random_state=rng)
    X = dd.from_array(X_, chunksize=10)
    y = dd.from_array(y_, chunksize=10)

    n_samples = n_workers - 1
    valid_X_, valid_y_ = make_classification(n_samples=n_samples, random_state=rng)
    valid_X = dd.from_array(valid_X_, chunksize=n_samples)
    valid_y = dd.from_array(valid_y_, chunksize=n_samples)

    cls = dxgb.DaskXGBClassifier(
        device=device, n_estimators=2, eval_metric=["auc", "aucpr"]
    )
    cls.fit(X, y, eval_set=[(valid_X, valid_y)])

    # multiclass
    X_, y_ = make_classification(
        n_samples=n_samples,
        n_classes=n_workers,
        n_informative=n_features,
        n_redundant=0,
        n_repeated=0,
    )
    for i in range(y_.shape[0]):
        y_[i] = i % n_workers
    X = dd.from_array(X_, chunksize=10)
    y = dd.from_array(y_, chunksize=10)

    n_samples = n_workers - 1
    valid_X_, valid_y_ = make_classification(
        n_samples=n_samples,
        n_classes=n_workers,
        n_informative=n_features,
        n_redundant=0,
        n_repeated=0,
    )
    for i in range(valid_y_.shape[0]):
        valid_y_[i] = i % n_workers
    valid_X = dd.from_array(valid_X_, chunksize=n_samples)
    valid_y = dd.from_array(valid_y_, chunksize=n_samples)

    # Specify base score in case if there are only two workers and one sample.
    cls = dxgb.DaskXGBClassifier(
        device=device, n_estimators=2, eval_metric=["auc", "aucpr"], base_score=0.5
    )
    cls.fit(X, y, eval_set=[(valid_X, valid_y)])


@pytest.mark.parametrize(
    "client_kwargs",
    [pytest.param({"n_workers": 4, "dashboard_address": ":0"}, id="4-workers")],
    indirect=True,
)
def test_empty_dmatrix_auc(client: "Client") -> None:
    run_empty_dmatrix_auc(client, "cpu", 4)


def run_auc(client: "Client", device: str) -> None:
    from sklearn import datasets

    n_samples = 100
    n_features = 97
    rng = np.random.RandomState(1994)
    X_, y_ = datasets.make_classification(
        n_samples=n_samples, n_features=n_features, random_state=rng
    )
    X = dd.from_array(X_, chunksize=10)
    y = dd.from_array(y_, chunksize=10)

    valid_X_, valid_y_ = datasets.make_classification(
        n_samples=n_samples, n_features=n_features, random_state=rng
    )
    valid_X = dd.from_array(valid_X_, chunksize=10)
    valid_y = dd.from_array(valid_y_, chunksize=10)

    cls = xgb.XGBClassifier(device=device, n_estimators=2, eval_metric="auc")
    cls.fit(X_, y_, eval_set=[(valid_X_, valid_y_)])

    dcls = dxgb.DaskXGBClassifier(device=device, n_estimators=2, eval_metric="auc")
    dcls.fit(X, y, eval_set=[(valid_X, valid_y)])

    approx = dcls.evals_result()["validation_0"]["auc"]
    exact = cls.evals_result()["validation_0"]["auc"]
    for i in range(2):
        # approximated test.
        assert np.abs(approx[i] - exact[i]) <= 0.06


def test_auc(client: "Client") -> None:
    run_auc(client, "cpu")


# No test for Exact, as empty DMatrix handling are mostly for distributed
# environment and Exact doesn't support it.
@pytest.mark.parametrize("tree_method", ["hist", "approx"])
def test_empty_dmatrix(tree_method: str, client: "Client") -> None:
    parameters = {"tree_method": tree_method}
    run_empty_dmatrix_reg(client, parameters)
    run_empty_dmatrix_cls(client, parameters)
    parameters = {"tree_method": tree_method, "objective": "reg:absoluteerror"}
    run_empty_dmatrix_reg(client, parameters)


async def run_from_dask_array_asyncio(scheduler_address: str) -> dxgb.TrainReturnT:
    async with Client(scheduler_address, asynchronous=True) as client:
        X, y, _ = generate_array()
        m = await DaskDMatrix(client, X, y)
        output = await dxgb.train(client, {}, dtrain=m)

        with_m = await dxgb.predict(client, output, m)
        with_X = await dxgb.predict(client, output, X)
        inplace = await dxgb.inplace_predict(client, output, X)
        assert isinstance(with_m, da.Array)
        assert isinstance(with_X, da.Array)
        assert isinstance(inplace, da.Array)

        np.testing.assert_allclose(
            await client.compute(with_m), await client.compute(with_X)
        )
        np.testing.assert_allclose(
            await client.compute(with_m), await client.compute(inplace)
        )
    return output


async def run_dask_regressor_asyncio(scheduler_address: str) -> None:
    async with Client(scheduler_address, asynchronous=True) as client:
        X, y, _ = generate_array()
        X = X[:200]
        y = y[:200]
        expected_rows = X.shape[0]
        regressor = await dxgb.DaskXGBRegressor(verbosity=1, n_estimators=2)
        regressor.set_params(tree_method="hist")
        regressor.client = client
        await regressor.fit(X, y, eval_set=[(X, y)])
        prediction = await regressor.predict(X)

        assert prediction.ndim == 1
        assert prediction.shape[0] == expected_rows

        history = regressor.evals_result()

        assert isinstance(prediction, da.Array)
        assert isinstance(history, dict)

        assert list(history["validation_0"].keys())[0] == "rmse"
        assert len(history["validation_0"]["rmse"]) == 2

        awaited = await client.compute(prediction)
        assert awaited.shape[0] == expected_rows


async def run_dask_classifier_asyncio(scheduler_address: str) -> None:
    async with Client(scheduler_address, asynchronous=True) as client:
        X, y, _ = generate_array()
        expected_rows = X.shape[0]
        y = (y * 10).astype(np.int32)
        classifier = await dxgb.DaskXGBClassifier(
            verbosity=1, n_estimators=2, eval_metric="merror"
        )
        classifier.client = client
        await classifier.fit(X, y, eval_set=[(X, y)])
        prediction = await classifier.predict(X)

        assert prediction.ndim == 1
        assert prediction.shape[0] == expected_rows

        history = classifier.evals_result()

        assert isinstance(prediction, da.Array)
        assert isinstance(history, dict)

        assert list(history.keys())[0] == "validation_0"
        assert list(history["validation_0"].keys())[0] == "merror"
        assert len(list(history["validation_0"])) == 1
        assert len(history["validation_0"]["merror"]) == 2

        # Test .predict_proba()
        probas = await classifier.predict_proba(X)
        assert classifier.n_classes_ == 10
        assert probas.ndim == 2
        assert probas.shape[0] == expected_rows
        assert probas.shape[1] == 10

        # Test with dataframe.
        X_d = dd.from_dask_array(X)
        y_d = dd.from_dask_array(y)
        await classifier.fit(X_d, y_d)

        assert classifier.n_classes_ == 10
        prediction = await client.compute(await classifier.predict(X_d))

        assert prediction.ndim == 1
        assert prediction.shape[0] == expected_rows


def test_with_asyncio(client: "Client") -> None:
    address = client.scheduler.address
    output = asyncio.run(run_from_dask_array_asyncio(address))
    assert isinstance(output["booster"], xgb.Booster)
    assert isinstance(output["history"], dict)

    asyncio.run(run_dask_regressor_asyncio(address))
    asyncio.run(run_dask_classifier_asyncio(address))


async def generate_concurrent_trainings() -> None:
    async def train() -> None:
        async with LocalCluster(
            n_workers=2, threads_per_worker=1, asynchronous=True, dashboard_address=":0"
        ) as cluster:
            async with Client(cluster, asynchronous=True) as client:
                X, y, w = generate_array(with_weights=True)
                dtrain = await DaskDMatrix(client, X, y, weight=w)
                dvalid = await DaskDMatrix(client, X, y, weight=w)
                output = await dxgb.train(client, {}, dtrain=dtrain)
                await dxgb.predict(client, output, data=dvalid)

    await asyncio.gather(train(), train())


def test_concurrent_trainings() -> None:
    asyncio.run(generate_concurrent_trainings())


def test_predict(client: "Client") -> None:
    X, y, _ = generate_array()
    dtrain = DaskDMatrix(client, X, y)
    booster = dxgb.train(client, {}, dtrain, num_boost_round=2)["booster"]

    predt_0 = dxgb.predict(client, model=booster, data=dtrain)
    assert predt_0.ndim == 1
    assert predt_0.shape[0] == kRows

    margin = dxgb.predict(client, model=booster, data=dtrain, output_margin=True)
    assert margin.ndim == 1
    assert margin.shape[0] == kRows

    shap = dxgb.predict(client, model=booster, data=dtrain, pred_contribs=True)
    assert shap.ndim == 2
    assert shap.shape[0] == kRows
    assert shap.shape[1] == kCols + 1

    booster_f = client.scatter(booster, broadcast=True)

    predt_1 = dxgb.predict(client, booster_f, X).compute()
    predt_2 = dxgb.inplace_predict(client, booster_f, X).compute()
    np.testing.assert_allclose(predt_0, predt_1)
    np.testing.assert_allclose(predt_0, predt_2)


def test_predict_with_meta(client: "Client") -> None:
    X, y, w = generate_array(with_weights=True)
    assert w is not None
    partition_size = 20
    margin = da.random.random(kRows, partition_size) + 1e4

    dtrain = DaskDMatrix(client, X, y, weight=w, base_margin=margin)
    booster: xgb.Booster = dxgb.train(client, {}, dtrain, num_boost_round=4)["booster"]

    prediction = dxgb.predict(client, model=booster, data=dtrain)
    assert prediction.ndim == 1
    assert prediction.shape[0] == kRows

    prediction = client.compute(prediction).result()
    assert np.all(prediction > 1e3)

    m = xgb.DMatrix(X.compute())
    m.set_info(label=y.compute(), weight=w.compute(), base_margin=margin.compute())
    single = booster.predict(m)  # Make sure the ordering is correct.
    assert np.all(prediction == single)


def run_aft_survival(client: "Client", dmatrix_t: Type) -> None:
    df = dd.read_csv(os.path.join(tm.data_dir(__file__), "veterans_lung_cancer.csv"))
    y_lower_bound = df["Survival_label_lower_bound"]
    y_upper_bound = df["Survival_label_upper_bound"]
    X = df.drop(["Survival_label_lower_bound", "Survival_label_upper_bound"], axis=1)
    m = dmatrix_t(
        client, X, label_lower_bound=y_lower_bound, label_upper_bound=y_upper_bound
    )
    base_params = {
        "verbosity": 0,
        "objective": "survival:aft",
        "eval_metric": "aft-nloglik",
        "learning_rate": 0.05,
        "aft_loss_distribution_scale": 1.20,
        "max_depth": 6,
        "lambda": 0.01,
        "alpha": 0.02,
    }

    nloglik_rec = {}
    dists = ["normal", "logistic", "extreme"]
    for dist in dists:
        params = base_params
        params.update({"aft_loss_distribution": dist})
        evals_result = {}
        out = dxgb.train(client, params, m, num_boost_round=100, evals=[(m, "train")])
        evals_result = out["history"]
        nloglik_rec[dist] = evals_result["train"]["aft-nloglik"]
        # AFT metric (negative log likelihood) improve monotonically
        assert all(p >= q for p, q in zip(nloglik_rec[dist], nloglik_rec[dist][:1]))
    # For this data, normal distribution works the best
    assert nloglik_rec["normal"][-1] < 4.9
    assert nloglik_rec["logistic"][-1] > 4.9
    assert nloglik_rec["extreme"][-1] > 4.9


def test_dask_aft_survival(client: "Client") -> None:
    run_aft_survival(client, DaskDMatrix)


@pytest.mark.parametrize("booster", ["dart", "gbtree"])
def test_dask_predict_leaf(booster: str, client: "Client") -> None:
    from sklearn.datasets import load_digits

    X_, y_ = load_digits(return_X_y=True)
    X_, _, y_, _ = train_test_split(
        X_, y_, train_size=300, stratify=y_, random_state=1994
    )
    num_parallel_tree = 4
    X, y = dd.from_array(X_, chunksize=32), dd.from_array(y_, chunksize=32)
    rounds = 4
    cls = dxgb.DaskXGBClassifier(
        n_estimators=rounds, num_parallel_tree=num_parallel_tree, booster=booster
    )
    cls.client = client
    cls.fit(X, y)
    leaf = dxgb.predict(
        client,
        cls.get_booster(),
        X.to_dask_array(),  # we can't map_blocks on dataframe when output is 4-dim.
        pred_leaf=True,
        strict_shape=True,
        validate_features=False,
    ).compute()

    assert leaf.shape[0] == X_.shape[0]
    assert leaf.shape[1] == rounds
    assert leaf.shape[2] == cls.n_classes_
    assert leaf.shape[3] == num_parallel_tree

    leaf_from_apply = cls.apply(X).reshape(leaf.shape).compute()
    np.testing.assert_allclose(leaf_from_apply, leaf)

    validate_leaf_output(leaf, num_parallel_tree)


def test_dask_iteration_range(client: "Client") -> None:
    X, y, _ = generate_array()
    n_rounds = 5

    Xy = xgb.DMatrix(X.compute(), y.compute())

    dXy = dxgb.DaskDMatrix(client, X, y)
    booster = dxgb.train(
        client, {"tree_method": "hist"}, dXy, num_boost_round=n_rounds
    )["booster"]

    for i in range(0, n_rounds):
        iter_range = (0, i)
        native_predt = booster.predict(Xy, iteration_range=iter_range)

        with_dask_dmatrix = dxgb.predict(
            client, booster, dXy, iteration_range=iter_range
        )
        with_dask_collection = dxgb.predict(
            client, booster, X, iteration_range=iter_range
        )
        with_inplace = dxgb.inplace_predict(
            client, booster, X, iteration_range=iter_range
        )
        np.testing.assert_allclose(native_predt, with_dask_dmatrix.compute())
        np.testing.assert_allclose(native_predt, with_dask_collection.compute())
        np.testing.assert_allclose(native_predt, with_inplace.compute())

    full_predt = dxgb.predict(client, booster, X, iteration_range=(0, n_rounds))
    default = dxgb.predict(client, booster, X)
    np.testing.assert_allclose(full_predt.compute(), default.compute())


def test_killed_task_wo_hang(client: "Client") -> None:
    # Test that aborting a worker doesn't lead to hang.
    class Eve(xgb.callback.TrainingCallback):
        def after_iteration(
            self, model: xgb.Booster, epoch: int, evals_log: Dict
        ) -> bool:
            if coll.get_rank() == 1:
                os.abort()
            return False

    X, y, _ = generate_array()
    n_rounds = 10
    dXy = dxgb.DaskDMatrix(client, X, y)
    # The precise error message depends on Dask scheduler.
    try:
        dxgb.train(
            client,
            {"tree_method": "hist"},
            dXy,
            num_boost_round=n_rounds,
            callbacks=[Eve()],
        )
    except (ValueError, KilledWorker):
        # These exceptions indicate that the killed worker caused training to fail
        # promptly, which is the expected behavior for this test. We only verify
        # that training does not hang, so the exceptions are intentionally ignored.
        pass


def test_invalid_config(client: "Client") -> None:
    X, y, _ = generate_array()
    dtrain = DaskDMatrix(client, X, y)

    with dask.config.set({"xgboost.foo": "bar"}):
        with pytest.raises(ValueError, match=r"Unknown configuration.*"):
            dxgb.train(client, {}, dtrain, num_boost_round=4)

    with dask.config.set({"xgboost.scheduler_address": "127.0.0.1:foo"}):
        with pytest.raises(socket.gaierror, match=r".*not known.*"):
            dxgb.train(client, {}, dtrain, num_boost_round=1)

    # No failure only because we are also using the Dask scheduler address.
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        port = s.getsockname()[1]
        cfg = CollConfig(tracker_host_ip="127.0.0.1", tracker_port=port)
        dxgb.train(client, {}, dtrain, num_boost_round=1, coll_cfg=cfg)

    with pytest.raises(ValueError, match=r"comm_group.*timeout >= 0.*"):
        cfg = CollConfig(tracker_host_ip="127.0.0.1", tracker_port=0, timeout=-1)
        dxgb.train(client, {}, dtrain, num_boost_round=1, coll_cfg=cfg)


def test_worker_port(client_one_worker: "Client") -> None:
    from xgboost.testing.collective import get_avail_port

    X, y, _ = generate_array()
    dtrain = DaskDMatrix(client_one_worker, X, y)

    cfg = CollConfig(worker_port=get_avail_port)
    dxgb.train(client_one_worker, {}, dtrain, num_boost_round=4, coll_cfg=cfg)


class TestWithDask:
    def test_dmatrix_binary(self, client: "Client", tmp_path: Path) -> None:
        def save_dmatrix(rabit_args: Dict[str, Union[int, str]], tmpdir: str) -> None:
            with dxgb.CommunicatorContext(**rabit_args):
                rank = xgb.collective.get_rank()
                X, y = tm.make_categorical(100, 4, 4, onehot=False)
                Xy = xgb.DMatrix(X, y)
                path = os.path.join(tmpdir, f"{rank}.bin")
                Xy.save_binary(path)

        def load_dmatrix(rabit_args: Dict[str, Union[int, str]], tmpdir: str) -> None:
            with dxgb.CommunicatorContext(**rabit_args):
                rank = xgb.collective.get_rank()
                path = os.path.join(tmpdir, f"{rank}.bin")
                Xy = xgb.DMatrix(path)
                assert Xy.num_row() == 100
                assert Xy.num_col() == 4

        workers = tm.dask.get_client_workers(client)
        rabit_args = get_rabit_args(client, len(workers))
        futures = []
        for w in workers:
            # same argument for each worker, must set pure to False otherwise dask
            # will try to reuse the result from the first worker and hang waiting
            # for it.
            f = client.submit(
                save_dmatrix, rabit_args, str(tmp_path), workers=[w], pure=False
            )
            futures.append(f)
        client.gather(futures)

        rabit_args = get_rabit_args(client, len(workers))
        futures = []
        for w in workers:
            f = client.submit(
                load_dmatrix, rabit_args, str(tmp_path), workers=[w], pure=False
            )
            futures.append(f)
        client.gather(futures)

    @pytest.mark.parametrize(
        "config_key,config_value", [("verbosity", 0), ("use_rmm", True)]
    )
    def test_global_config(
        self, client: "Client", config_key: str, config_value: Any
    ) -> None:
        X, y, _ = generate_array()
        xgb.config.set_config(**{config_key: config_value})
        dtrain = DaskDMatrix(client, X, y)
        before_fname = "./before_training-test_global_config"
        after_fname = "./after_training-test_global_config"

        class TestCallback(xgb.callback.TrainingCallback):
            def write_file(self, fname: str) -> None:
                with open(fname, "w") as fd:
                    fd.write(str(xgb.config.get_config()[config_key]))

            def before_training(self, model: xgb.Booster) -> xgb.Booster:
                self.write_file(before_fname)
                assert xgb.config.get_config()[config_key] == config_value
                return model

            def after_training(self, model: xgb.Booster) -> xgb.Booster:
                assert xgb.config.get_config()[config_key] == config_value
                return model

            def before_iteration(
                self, model: xgb.Booster, epoch: int, evals_log: Dict
            ) -> bool:
                assert xgb.config.get_config()[config_key] == config_value
                return False

            def after_iteration(
                self, model: xgb.Booster, epoch: int, evals_log: Dict
            ) -> bool:
                self.write_file(after_fname)
                assert xgb.config.get_config()[config_key] == config_value
                return False

        dxgb.train(client, {}, dtrain, num_boost_round=4, callbacks=[TestCallback()])[
            "booster"
        ]

        with open(before_fname, "r") as before, open(after_fname, "r") as after:
            assert before.read() == str(config_value)
            assert after.read() == str(config_value)

        os.remove(before_fname)
        os.remove(after_fname)

    def run_updater_test(
        self,
        client: "Client",
        params: Dict,
        num_rounds: int,
        dataset: tm.TestDataset,
        tree_method: str,
    ) -> None:
        params["tree_method"] = tree_method
        params["debug_synchronize"] = True
        params = dataset.set_params(params)

        # It doesn't make sense to distribute a completely empty dataset.
        assume(dataset.X.shape[0] != 0)

        chunk = 128
        y_chunk = chunk if len(dataset.y.shape) == 1 else (chunk, dataset.y.shape[1])
        X = da.from_array(dataset.X, chunks=(chunk, dataset.X.shape[1]))
        y = da.from_array(dataset.y, chunks=y_chunk)
        if dataset.w is not None:
            w = da.from_array(dataset.w, chunks=(chunk,))
        else:
            w = None

        m = dxgb.DaskDMatrix(client, data=X, label=y, weight=w)
        history = dxgb.train(
            client,
            params=params,
            dtrain=m,
            num_boost_round=num_rounds,
            evals=[(m, "train")],
        )["history"]
        note(str(history))
        history = history["train"][dataset.metric]

        def is_stump() -> bool:
            return (
                params.get("max_depth", None) == 1
                or params.get("max_leaves", None) == 1
            )

        def minimum_bin() -> bool:
            return "max_bin" in params and params["max_bin"] == 2

        # See note on `ObjFunction::UpdateTreeLeaf`.
        update_leaf = dataset.name.endswith("-l1")
        if update_leaf and (is_stump() or minimum_bin()):
            assert tm.non_increasing(history, tolerance=1e-2)
            return
        elif minimum_bin() and is_stump():
            assert tm.non_increasing(history, tolerance=1e-3)
        else:
            assert tm.non_increasing(history)
        # Make sure that it's decreasing
        if is_stump():
            # we might have already got the best score with base_score.
            assert history[-1] <= history[0] + 1e-3
        else:
            assert history[-1] < history[0]

    @given(
        params=hist_parameter_strategy,
        cache_param=hist_cache_strategy,
        dataset=tm.make_dataset_strategy(),
    )
    @settings(
        deadline=None, max_examples=3, suppress_health_check=suppress, print_blob=True
    )
    def test_hist(
        self,
        params: Dict[str, Any],
        cache_param: Dict[str, Any],
        dataset: tm.TestDataset,
        client: "Client",
    ) -> None:
        num_rounds = 10
        params.update(cache_param)
        self.run_updater_test(client, params, num_rounds, dataset, "hist")

    def test_quantile_dmatrix(self, client: Client) -> None:
        X, y = make_categorical(client, 3000, 30, 13)

        Xy = dxgb.DaskDMatrix(client, X, y)
        valid_Xy = dxgb.DaskDMatrix(client, X, y)

        output = dxgb.train(
            client,
            {"tree_method": "hist"},
            Xy,
            num_boost_round=10,
            evals=[(Xy, "Train"), (valid_Xy, "Valid")],
        )
        dmatrix_hist = output["history"]

        Xy = dxgb.DaskQuantileDMatrix(client, X, y)
        valid_Xy = dxgb.DaskQuantileDMatrix(client, X, y, ref=Xy)

        output = dxgb.train(
            client,
            {"tree_method": "hist"},
            Xy,
            num_boost_round=10,
            evals=[(Xy, "Train"), (valid_Xy, "Valid")],
        )
        quantile_hist = output["history"]

        np.testing.assert_allclose(
            quantile_hist["Train"]["rmse"], dmatrix_hist["Train"]["rmse"]
        )
        np.testing.assert_allclose(
            quantile_hist["Valid"]["rmse"], dmatrix_hist["Valid"]["rmse"]
        )

    def test_empty_quantile_dmatrix(self, client: Client) -> None:
        X, y = make_categorical(client, 1, 16, 4, onehot=True)
        X_valid, y_valid = make_categorical(client, 2000, 16, 4, onehot=True)

        Xy = dxgb.DaskQuantileDMatrix(client, X, y)
        Xy_valid = dxgb.DaskQuantileDMatrix(client, X_valid, y_valid, ref=Xy)
        result = dxgb.train(
            client,
            {"tree_method": "hist"},
            Xy,
            num_boost_round=10,
            evals=[(Xy_valid, "Valid")],
        )
        predt = dxgb.inplace_predict(client, result["booster"], X).compute()
        np.testing.assert_allclose(y.compute(), predt)
        rmse = result["history"]["Valid"]["rmse"][-1]
        assert rmse < 6.5

    @given(
        params=hist_parameter_strategy,
        cache_param=hist_cache_strategy,
        dataset=tm.make_dataset_strategy(),
    )
    @settings(
        deadline=None, max_examples=10, suppress_health_check=suppress, print_blob=True
    )
    def test_approx(
        self,
        client: "Client",
        params: Dict,
        cache_param: Dict[str, Any],
        dataset: tm.TestDataset,
    ) -> None:
        num_rounds = 10
        params.update(cache_param)
        self.run_updater_test(client, params, num_rounds, dataset, "approx")

    def test_adaptive(self, client: "Client") -> None:
        def local_test(rabit_args: Dict[str, Union[int, str]], worker_id: int) -> bool:
            with dxgb.CommunicatorContext(**rabit_args):
                if worker_id == 0:
                    y = np.array([0.0, 0.0, 0.0])
                    x = np.array([[0.0]] * 3)
                else:
                    y = np.array([1000.0])
                    x = np.array(
                        [
                            [0.0],
                        ]
                    )

                Xy = xgb.DMatrix(x, y)
                booster = xgb.train(
                    {"tree_method": "hist", "objective": "reg:absoluteerror"},
                    Xy,
                    num_boost_round=1,
                )
                config = json.loads(booster.save_config())
                base_score = get_basescore(config)
                assert base_score == [250.0]
                return True

        workers = tm.dask.get_client_workers(client)
        rabit_args = get_rabit_args(client, len(workers))
        futures = []
        for i, _ in enumerate(workers):
            f = client.submit(local_test, rabit_args, i)
            futures.append(f)

        results = client.gather(futures)
        assert all(results)

    def test_n_workers(self, client: "Client") -> None:
        """Check obtaining worker addresses using input data."""

        def from_delayed(fut: Any, x: np.ndarray) -> Any:
            return da.from_delayed(fut, shape=x.shape, dtype=x.dtype)

        def place_on_worker(data: np.ndarray, worker: str) -> Any:
            return client.submit(
                lambda x: x,
                data,
                workers=[worker],
                allow_other_workers=False,
                pure=False,
            )

        workers = tm.dask.get_client_workers(client)
        from sklearn.datasets import load_breast_cancer

        X, y = load_breast_cancer(return_X_y=True)

        # Use client.submit to place data on specific workers with unique keys.
        X_fut_0 = place_on_worker(X, workers[0])
        y_fut_0 = place_on_worker(y, workers[0])
        distributed.wait([X_fut_0, y_fut_0])
        dX = from_delayed(X_fut_0, X)
        dy = from_delayed(y_fut_0, y)
        train = dxgb.DaskDMatrix(client, dX, dy)

        X_fut_1 = place_on_worker(X, workers[1])
        y_fut_1 = place_on_worker(y, workers[1])
        distributed.wait([X_fut_1, y_fut_1])
        dX_valid = from_delayed(X_fut_1, X)
        dy_valid = from_delayed(y_fut_1, y)
        valid = dxgb.DaskDMatrix(client, dX_valid, dy_valid)

        merged = dxgb._get_workers_from_data(train, evals=[(valid, "Valid")])
        assert len(merged) == 2

    @pytest.mark.skipif(**tm.no_dask())
    def test_feature_weights(self, client: "Client") -> None:
        kRows = 1024
        kCols = 64
        rng = da.random.RandomState(1994)
        X = rng.random_sample((kRows, kCols), chunks=(32, -1))
        y = rng.random_sample(kRows, chunks=32)

        fw = np.ones(shape=(kCols,))
        for i in range(kCols):
            fw[i] *= float(i)
        fw = da.from_array(fw)
        parser = os.path.join(tm.demo_dir(__file__), "guide-python", "model_parser.py")
        poly_increasing = get_feature_weights(
            X=X,
            y=y,
            fw=fw,
            parser_path=parser,
            tree_method="approx",
            model=dxgb.DaskXGBRegressor,
        )

        fw = np.ones(shape=(kCols,))
        for i in range(kCols):
            fw[i] *= float(kCols - i)
        fw = da.from_array(fw)
        poly_decreasing = get_feature_weights(
            X=X,
            y=y,
            fw=fw,
            parser_path=parser,
            tree_method="approx",
            model=dxgb.DaskXGBRegressor,
        )

        # Approxmated test, this is dependent on the implementation of random
        # number generator in std library.
        assert poly_increasing[0] > 0.08
        assert poly_decreasing[0] < -0.08

    @pytest.mark.skipif(**tm.no_dask())
    @pytest.mark.skipif(**tm.no_sklearn())
    def test_custom_objective(self, client: "Client", tmp_path: Path) -> None:
        X, y = get_california_housing()
        X, y = da.from_array(X), da.from_array(y)
        rounds = 20

        path = tmp_path / "log"

        def sqr(
            labels: np.ndarray, predts: np.ndarray
        ) -> Tuple[np.ndarray, np.ndarray]:
            with open(path, "a") as fd:
                print("Running sqr", file=fd)
            grad = predts - labels
            hess = np.ones(shape=labels.shape[0])
            return grad, hess

        reg = dxgb.DaskXGBRegressor(
            n_estimators=rounds, objective=sqr, tree_method="hist"
        )
        reg.fit(X, y, eval_set=[(X, y)])

        # Check the obj is ran for rounds.
        with open(path, "r") as fd:
            out = fd.readlines()
            assert len(out) == rounds

        results_custom = reg.evals_result()

        reg = dxgb.DaskXGBRegressor(
            n_estimators=rounds, tree_method="hist", base_score=0.5
        )
        reg.fit(X, y, eval_set=[(X, y)])
        results_native = reg.evals_result()

        np.testing.assert_allclose(
            results_custom["validation_0"]["rmse"],
            results_native["validation_0"]["rmse"],
        )
        tm.non_increasing(results_native["validation_0"]["rmse"])

        reg = dxgb.DaskXGBRegressor(
            n_estimators=rounds, objective=tm.ls_obj, tree_method="hist"
        )
        rng = da.random.RandomState(1994)
        w = rng.uniform(low=0.0, high=1.0, size=y.shape[0])
        reg.fit(X, y, sample_weight=w, eval_set=[(X, y)], sample_weight_eval_set=[w])
        results_custom = reg.evals_result()
        tm.non_increasing(results_custom["validation_0"]["rmse"])

    @pytest.mark.skipif(**tm.no_sklearn())
    def test_custom_metrics(self, client: "Client") -> None:
        from sklearn.datasets import make_classification
        from sklearn.metrics import hamming_loss, hinge_loss, log_loss

        Xn, yn = make_classification(random_state=2025)
        X, y = da.array(Xn), da.array(yn)

        clf = dxgb.DaskXGBClassifier(
            eval_metric=["logloss", hinge_loss], n_estimators=2
        )
        clf.fit(X, y, eval_set=[(X, y)])
        results = clf.evals_result()["validation_0"]
        assert "logloss" in results
        assert "hinge_loss" in results

        clf = dxgb.DaskXGBClassifier(
            eval_metric=[hamming_loss, log_loss], n_estimators=2
        )
        with pytest.raises(
            NotImplementedError, match="multiple custom metrics is not yet supported."
        ):
            clf.fit(X, y, eval_set=[(X, y)])

    def test_no_duplicated_partition(self) -> None:
        """Assert each worker has the correct amount of data, and DMatrix initialization
        doesn't generate unnecessary copies of data.

        """
        n_threads = os.cpu_count()
        assert n_threads is not None
        # This test needs a fresh client
        with LocalCluster(
            n_workers=2,
            threads_per_worker=max(1, n_threads // 2),
            dashboard_address=":0",
        ) as cluster:
            with Client(cluster) as client:
                X, y, _ = generate_array()
                n_partitions = X.npartitions
                m = dxgb.DaskDMatrix(client, X, y)
                workers = tm.dask.get_client_workers(client)
                rabit_args = get_rabit_args(client, len(workers))
                n_workers = len(workers)

                def worker_fn(worker_addr: str, data_ref: Dict) -> None:
                    from xgboost.dask.data import _dmatrix_from_list_of_parts

                    with dxgb.CommunicatorContext(**rabit_args):
                        local_dtrain = _dmatrix_from_list_of_parts(
                            **data_ref,
                            nthread=7,
                            model=None,
                            Xy_cats=None,
                        )
                        total = np.array([local_dtrain.num_row()])
                        total = xgb.collective.allreduce(total, xgb.collective.Op.SUM)
                        assert total[0] == kRows

                futures = []
                for i in range(len(workers)):
                    futures.append(
                        client.submit(
                            worker_fn,
                            workers[i],
                            m._create_fn_args(workers[i]),
                            pure=False,
                            workers=[workers[i]],
                        )
                    )
                client.gather(futures)

                has_what = client.has_what()
                cnt = 0
                data = set()
                for k, v in has_what.items():
                    for d in v:
                        cnt += 1
                        data.add(d)

                assert len(data) == cnt
                # Subtract the on disk resource from each worker
                assert cnt - n_workers == n_partitions

    def test_data_initialization(self, client: "Client") -> None:
        """assert that we don't create duplicated DMatrix"""
        from sklearn.datasets import load_digits

        X, y = load_digits(return_X_y=True)
        X, _, y, _ = train_test_split(
            X, y, train_size=200, stratify=y, random_state=1994
        )
        X, y = dd.from_array(X, chunksize=32), dd.from_array(y, chunksize=32)
        validate_data_initialization(
            dxgb.DaskQuantileDMatrix, dxgb.DaskXGBClassifier, X, y
        )

    def run_shap(
        self, X: Any, y: Any, params: Dict[str, Any], client: "Client"
    ) -> None:
        rows = X.shape[0]
        cols = X.shape[1]

        def assert_shape(shape: Tuple[int, ...]) -> None:
            assert shape[0] == rows
            if "num_class" in params.keys():
                assert shape[1] == params["num_class"]
                assert shape[2] == cols + 1
            else:
                assert shape[1] == cols + 1

        X, y = da.from_array(X, chunks=(32, -1)), da.from_array(y, chunks=32)
        Xy = dxgb.DaskDMatrix(client, X, y)
        booster = dxgb.train(client, params, Xy, num_boost_round=10)["booster"]

        test_Xy = dxgb.DaskDMatrix(client, X, y)

        shap = dxgb.predict(client, booster, test_Xy, pred_contribs=True).compute()
        margin = dxgb.predict(client, booster, test_Xy, output_margin=True).compute()
        assert_shape(shap.shape)
        assert np.allclose(np.sum(shap, axis=len(shap.shape) - 1), margin, 1e-5, 1e-5)

        shap = dxgb.predict(client, booster, X, pred_contribs=True).compute()
        margin = dxgb.predict(client, booster, X, output_margin=True).compute()
        assert_shape(shap.shape)
        assert np.allclose(np.sum(shap, axis=len(shap.shape) - 1), margin, 1e-5, 1e-5)

        if "num_class" not in params.keys():
            X = dd.from_dask_array(X).repartition(npartitions=32)
            y = dd.from_dask_array(y).repartition(npartitions=32)
            shap_df = dxgb.predict(
                client, booster, X, pred_contribs=True, validate_features=False
            ).compute()
            assert_shape(shap_df.shape)
            assert np.allclose(
                np.sum(shap_df, axis=len(shap_df.shape) - 1), margin, 1e-5, 1e-5
            )

    def run_shap_cls_sklearn(self, X: Any, y: Any, client: "Client") -> None:
        X, y = da.from_array(X, chunks=(32, -1)), da.from_array(y, chunks=32)
        cls = dxgb.DaskXGBClassifier(n_estimators=4)
        cls.client = client
        cls.fit(X, y)
        booster = cls.get_booster()

        test_Xy = dxgb.DaskDMatrix(client, X, y)

        shap = dxgb.predict(client, booster, test_Xy, pred_contribs=True).compute()
        margin = dxgb.predict(client, booster, test_Xy, output_margin=True).compute()
        assert np.allclose(np.sum(shap, axis=len(shap.shape) - 1), margin, 1e-5, 1e-5)

        shap = dxgb.predict(client, booster, X, pred_contribs=True).compute()
        margin = dxgb.predict(client, booster, X, output_margin=True).compute()
        assert np.allclose(np.sum(shap, axis=len(shap.shape) - 1), margin, 1e-5, 1e-5)

    def test_shap(self, client: "Client") -> None:
        from sklearn.datasets import load_diabetes, load_iris

        X, y = load_diabetes(return_X_y=True)
        params: Dict[str, Any] = {"objective": "reg:squarederror"}
        self.run_shap(X, y, params, client)

        X, y = load_iris(return_X_y=True)
        params = {"objective": "multi:softmax", "num_class": 3}
        self.run_shap(X, y, params, client)

        params = {"objective": "multi:softprob", "num_class": 3}
        self.run_shap(X, y, params, client)

        self.run_shap_cls_sklearn(X, y, client)

    def run_shap_interactions(
        self, X: Any, y: Any, params: Dict[str, Any], client: "Client"
    ) -> None:
        rows = X.shape[0]
        cols = X.shape[1]
        X, y = da.from_array(X, chunks=(32, -1)), da.from_array(y, chunks=32)

        Xy = dxgb.DaskDMatrix(client, X, y)
        booster = dxgb.train(client, params, Xy, num_boost_round=10)["booster"]

        test_Xy = dxgb.DaskDMatrix(client, X, y)

        shap = dxgb.predict(client, booster, test_Xy, pred_interactions=True).compute()

        assert len(shap.shape) == 3
        assert shap.shape[0] == rows
        assert shap.shape[1] == cols + 1
        assert shap.shape[2] == cols + 1

        margin = dxgb.predict(client, booster, test_Xy, output_margin=True).compute()
        assert np.allclose(
            np.sum(shap, axis=(len(shap.shape) - 1, len(shap.shape) - 2)),
            margin,
            1e-5,
            1e-5,
        )

    def test_shap_interactions(self, client: "Client") -> None:
        from sklearn.datasets import load_diabetes

        X, y = load_diabetes(return_X_y=True)
        params = {"objective": "reg:squarederror"}
        self.run_shap_interactions(X, y, params, client)

    @pytest.mark.skipif(**tm.no_sklearn())
    def test_sklearn_io(self, client: "Client", tmp_path: Path) -> None:
        from sklearn.datasets import load_digits

        X_, y_ = load_digits(return_X_y=True)
        X, y = da.from_array(X_), da.from_array(y_)
        cls = dxgb.DaskXGBClassifier(n_estimators=10)
        cls.client = client
        cls.fit(X, y)
        predt_0 = cls.predict(X)
        proba_0 = cls.predict_proba(X)

        path = tmp_path / "model.pkl"
        with open(path, "wb") as fd:
            pickle.dump(cls, fd)

        with open(path, "rb") as fd:
            cls = pickle.load(fd)
        predt_1 = cls.predict(X)
        proba_1 = cls.predict_proba(X)
        np.testing.assert_allclose(predt_0.compute(), predt_1.compute())
        np.testing.assert_allclose(proba_0.compute(), proba_1.compute())

        path = tmp_path / "cls.json"
        cls.save_model(path)

        cls = dxgb.DaskXGBClassifier()
        cls.load_model(path)
        assert cls.n_classes_ == 10
        predt_2 = cls.predict(X)
        proba_2 = cls.predict_proba(X)

        np.testing.assert_allclose(predt_0.compute(), predt_2.compute())
        np.testing.assert_allclose(proba_0.compute(), proba_2.compute())

        # Use single node to load
        cls = xgb.XGBClassifier()
        cls.load_model(path)
        assert cls.n_classes_ == 10
        predt_3 = cls.predict(X_)
        proba_3 = cls.predict_proba(X_)

        np.testing.assert_allclose(predt_0.compute(), predt_3)
        np.testing.assert_allclose(proba_0.compute(), proba_3)


def test_dask_unsupported_features(client: "Client") -> None:
    X, y, _ = generate_array()
    # gblinear doesn't support distributed training.
    with pytest.raises(NotImplementedError, match="gblinear"):
        dxgb.train(client, {"booster": "gblinear"}, dxgb.DaskDMatrix(client, X, y))


def test_parallel_submits(client: "Client") -> None:
    """Test for running multiple train simultaneously from single clients."""
    try:
        from distributed import MultiLock  # NOQA
    except ImportError:
        pytest.skip("`distributed.MultiLock' is not available")

    from sklearn.datasets import load_digits

    futures = []
    workers = tm.dask.get_client_workers(client)
    n_submits = len(workers)
    for i in range(n_submits):
        X_, y_ = load_digits(return_X_y=True)
        X_, _, y_, _ = train_test_split(
            X_, y_, train_size=300, stratify=y_, random_state=1994
        )
        X = dd.from_array(X_, chunksize=32)
        y = dd.from_array(y_, chunksize=32)
        cls = dxgb.DaskXGBClassifier(
            verbosity=1,
            n_estimators=i + 1,
            eval_metric="merror",
        )
        f = client.submit(cls.fit, X, y, pure=False)
        futures.append(f)

    classifiers = client.gather(futures)
    assert len(classifiers) == n_submits
    for i, cls in enumerate(classifiers):
        assert cls.get_booster().num_boosted_rounds() == i + 1


def run_tree_stats(client: Client, tree_method: str, device: str) -> str:
    """assert that different workers count dosn't affect summ statistic's on root"""

    def dask_train(
        X: np.ndarray, y: np.ndarray, num_obs: int, num_features: int
    ) -> Dict[str, Any]:
        chunk_size = 100
        X = da.from_array(X, chunks=(chunk_size, num_features))
        y = da.from_array(y.reshape(num_obs, 1), chunks=(chunk_size, 1))
        dtrain = dxgb.DaskDMatrix(client, X, y)

        output = dxgb.train(
            client,
            {
                "verbosity": 0,
                "tree_method": tree_method,
                "device": device,
                "objective": "reg:squarederror",
                "max_depth": 3,
            },
            dtrain,
            num_boost_round=1,
        )
        dump_model = output["booster"].get_dump(with_stats=True, dump_format="json")[0]
        return json.loads(dump_model)

    num_obs = 1000
    num_features = 10
    X, y = make_regression(num_obs, num_features, random_state=777)
    model = dask_train(X, y, num_obs, num_features)

    # asserts children have correct cover.
    stack = [model]
    while stack:
        node: dict = stack.pop()
        if "leaf" in node.keys():
            continue
        cover = 0
        for c in node["children"]:
            cover += c["cover"]
            stack.append(c)
        assert cover == node["cover"]

    return model["cover"]


@pytest.mark.parametrize("tree_method", ["hist", "approx"])
def test_tree_stats(
    tree_method: str, client_one_worker: "Client", client: "Client"
) -> None:
    with client_one_worker.as_current():
        local = run_tree_stats(client_one_worker, tree_method, "cpu")
    with client.as_current():
        distributed = run_tree_stats(client, tree_method, "cpu")

    assert local == distributed


def test_parallel_submit_multi_clients() -> None:
    """Test for running multiple train simultaneously from multiple clients."""
    try:
        from distributed import MultiLock  # NOQA
    except ImportError:
        pytest.skip("`distributed.MultiLock' is not available")

    from sklearn.datasets import load_digits

    with LocalCluster(n_workers=4, dashboard_address=":0") as cluster:
        with Client(cluster) as client:
            workers = tm.dask.get_client_workers(client)

        n_submits = len(workers)
        assert n_submits == 4
        futures = []

        for i in range(n_submits):
            client = Client(cluster)
            X_, y_ = load_digits(return_X_y=True)
            X_ += 1.0
            X = dd.from_array(X_, chunksize=32)
            y = dd.from_array(y_, chunksize=32)
            cls = dxgb.DaskXGBClassifier(
                verbosity=1,
                n_estimators=i + 1,
                eval_metric="merror",
            )
            f = client.submit(cls.fit, X, y, pure=False)
            futures.append((client, f))

        t_futures = []
        with ThreadPoolExecutor(max_workers=16) as e:
            for i in range(n_submits):

                def _() -> dxgb.DaskXGBClassifier:
                    return futures[i][0].compute(futures[i][1]).result()

                tf = e.submit(_)
                t_futures.append(tf)

        for i, tf in enumerate(t_futures):
            assert tf.result().get_booster().num_boosted_rounds() == i + 1


def test_init_estimation(client: Client) -> None:
    check_init_estimation("hist", "cpu", client)


@pytest.mark.parametrize("tree_method", ["hist", "approx"])
def test_uneven_nan(tree_method: str, client: "Client") -> None:
    n_workers = 2
    check_uneven_nan(client, tree_method, "cpu", n_workers)


class TestDaskCallbacks:
    @pytest.mark.skipif(**tm.no_sklearn())
    def test_early_stopping(self, client: "Client") -> None:
        from sklearn.datasets import load_breast_cancer

        X, y = load_breast_cancer(return_X_y=True)
        X, y = da.from_array(X), da.from_array(y)
        m = dxgb.DaskDMatrix(client, X, y)

        valid = dxgb.DaskDMatrix(client, X, y)
        early_stopping_rounds = 5
        booster = dxgb.train(
            client,
            {
                "objective": "binary:logistic",
                "eval_metric": "error",
                "tree_method": "hist",
            },
            m,
            evals=[(valid, "Valid")],
            num_boost_round=1000,
            early_stopping_rounds=early_stopping_rounds,
        )["booster"]
        assert hasattr(booster, "best_score")
        dump = booster.get_dump(dump_format="json")
        assert len(dump) - booster.best_iteration == early_stopping_rounds + 1

        valid_X, valid_y = load_breast_cancer(return_X_y=True)
        valid_X, valid_y = da.from_array(valid_X), da.from_array(valid_y)
        cls = dxgb.DaskXGBClassifier(
            objective="binary:logistic",
            tree_method="hist",
            n_estimators=1000,
            early_stopping_rounds=early_stopping_rounds,
        )
        cls.client = client
        cls.fit(
            X,
            y,
            eval_set=[(valid_X, valid_y)],
        )
        booster = cls.get_booster()
        dump = booster.get_dump(dump_format="json")
        assert len(dump) - booster.best_iteration == early_stopping_rounds + 1

        # Specify the metric
        cls = dxgb.DaskXGBClassifier(
            objective="binary:logistic",
            tree_method="hist",
            n_estimators=1000,
            early_stopping_rounds=early_stopping_rounds,
            eval_metric="error",
        )
        cls.client = client
        cls.fit(
            X,
            y,
            eval_set=[(valid_X, valid_y)],
        )
        assert tm.non_increasing(cls.evals_result()["validation_0"]["error"])
        booster = cls.get_booster()
        dump = booster.get_dump(dump_format="json")
        assert len(cls.evals_result()["validation_0"]["error"]) < 20
        assert len(dump) - booster.best_iteration == early_stopping_rounds + 1

    @pytest.mark.skipif(**tm.no_sklearn())
    def test_early_stopping_custom_eval(self, client: "Client") -> None:
        from sklearn.datasets import load_breast_cancer

        X, y = load_breast_cancer(return_X_y=True)
        X, y = da.from_array(X), da.from_array(y)
        m = dxgb.DaskDMatrix(client, X, y)

        def eval_error_metric(
            predt: np.ndarray, dtrain: xgb.DMatrix
        ) -> Tuple[str, np.float64]:
            return tm.eval_error_metric(predt, dtrain, rev_link=False)

        valid = dxgb.DaskDMatrix(client, X, y)
        early_stopping_rounds = 5
        booster = dxgb.train(
            client,
            {
                "objective": "binary:logistic",
                "eval_metric": "error",
                "tree_method": "hist",
            },
            m,
            evals=[(m, "Train"), (valid, "Valid")],
            custom_metric=eval_error_metric,
            num_boost_round=1000,
            early_stopping_rounds=early_stopping_rounds,
        )["booster"]
        assert hasattr(booster, "best_score")
        dump = booster.get_dump(dump_format="json")
        assert len(dump) - booster.best_iteration == early_stopping_rounds + 1

        valid_X, valid_y = load_breast_cancer(return_X_y=True)
        valid_X, valid_y = da.from_array(valid_X), da.from_array(valid_y)
        cls = dxgb.DaskXGBClassifier(
            objective="binary:logistic",
            tree_method="hist",
            n_estimators=1000,
            eval_metric=tm.eval_error_metric_skl,
            early_stopping_rounds=early_stopping_rounds,
        )
        cls.client = client
        cls.fit(
            X,
            y,
            eval_set=[(valid_X, valid_y)],
        )
        booster = cls.get_booster()
        dump = booster.get_dump(dump_format="json")
        assert len(dump) - booster.best_iteration == early_stopping_rounds + 1

    @pytest.mark.skipif(**tm.no_sklearn())
    def test_callback(self, client: "Client", tmp_path: Path) -> None:
        from sklearn.datasets import load_breast_cancer

        X, y = load_breast_cancer(return_X_y=True)
        X, y = da.from_array(X), da.from_array(y)

        cls = dxgb.DaskXGBClassifier(
            objective="binary:logistic",
            tree_method="hist",
            n_estimators=10,
            callbacks=[
                xgb.callback.TrainingCheckPoint(
                    directory=tmp_path, interval=1, name="model"
                )
            ],
        )
        cls.client = client
        cls.fit(
            X,
            y,
        )
        for i in range(1, 10):
            assert (
                tmp_path / f"model_{i}.{xgb.callback.TrainingCheckPoint.default_format}"
            ).exists()


@gen_cluster(
    client=True,
    clean_kwargs={"processes": False, "threads": False},
    allow_unclosed=True,
)
@pytest.mark.skip(reason="dmlc/xgboost#11405: test_worker_left is flaky")
async def test_worker_left(c: Client, s: Scheduler, a: Worker, b: Worker) -> None:
    async with Worker(s.address):
        dx = da.random.random((1000, 10)).rechunk(chunks=(10, None))
        dy = da.random.random((1000,)).rechunk(chunks=(10,))
        d_train = await dxgb.DaskDMatrix(
            c,
            dx,
            dy,
        )
    await async_poll_for(lambda: len(s.workers) == 2, timeout=5)
    with pytest.raises(RuntimeError, match="Missing"):
        await dxgb.train(
            c,
            {},
            d_train,
            evals=[(d_train, "train")],
        )


@gen_cluster(
    client=True,
    Worker=Nanny,
    clean_kwargs={"processes": False, "threads": False},
    allow_unclosed=True,
)
@pytest.mark.skip
async def test_worker_restarted(c: Client, s: Scheduler, a: Nanny, b: Nanny) -> None:
    dx = da.random.random((1000, 10)).rechunk(chunks=(10, None))
    dy = da.random.random((1000,)).rechunk(chunks=(10,))
    d_train = await dxgb.DaskDMatrix(
        c,
        dx,
        dy,
    )
    await c.restart_workers([a.worker_address])
    with pytest.raises(RuntimeError, match="Missing"):
        await dxgb.train(
            c,
            {},
            d_train,
            evals=[(d_train, "train")],
        )


def test_doc_link() -> None:
    for est in [
        dxgb.DaskXGBRegressor(),
        dxgb.DaskXGBClassifier(),
        dxgb.DaskXGBRanker(),
        dxgb.DaskXGBRFRegressor(),
        dxgb.DaskXGBRFClassifier(),
    ]:
        name = est.__class__.__name__
        link = est._get_doc_link()
        assert f"xgboost.dask.{name}" in link



"""Utilities for the CI."""

import os
from datetime import datetime, timedelta
from functools import wraps
from typing import Any, Callable, Dict, TypedDict, TypeVar, Union


class DirectoryExcursion:
    def __init__(self, path: Union[os.PathLike, str]) -> None:
        self.path = path
        self.curdir = os.path.normpath(os.path.abspath(os.path.curdir))

    def __enter__(self) -> None:
        os.chdir(self.path)

    def __exit__(self, *args: Any) -> None:
        os.chdir(self.curdir)


R = TypeVar("R")


def cd(path: Union[os.PathLike, str]) -> Callable:
    """Decorator for changing directory temporarily."""

    def chdir(func: Callable[..., R]) -> Callable[..., R]:
        @wraps(func)
        def inner(*args: Any, **kwargs: Any) -> R:
            with DirectoryExcursion(path):
                return func(*args, **kwargs)

        return inner

    return chdir


Record = TypedDict("Record", {"count": int, "total": timedelta})
timer: Dict[str, Record] = {}


def record_time(func: Callable[..., R]) -> Callable[..., R]:
    """Decorator for recording function runtime."""
    global timer

    @wraps(func)
    def inner(*args: Any, **kwargs: Any) -> R:
        if func.__name__ not in timer:
            timer[func.__name__] = {"count": 0, "total": timedelta(0)}
        s = datetime.now()
        try:
            r = func(*args, **kwargs)
        finally:
            e = datetime.now()
            timer[func.__name__]["count"] += 1
            timer[func.__name__]["total"] += e - s
        return r

    return inner


def print_time() -> None:
    """Print all recorded items by :py:func:`record_time`."""
    global timer
    for k, v in timer.items():
        print(
            "Name:",
            k,
            "Called:",
            v["count"],
            "Elapsed:",
            f"{v['total'].seconds} secs",
        )


ROOT = os.path.normpath(
    os.path.join(
        os.path.dirname(os.path.abspath(__file__)), os.path.pardir, os.path.pardir
    )
)
R_PACKAGE = os.path.join(ROOT, "R-package")
JVM_PACKAGES = os.path.join(ROOT, "jvm-packages")
PY_PACKAGE = os.path.join(ROOT, "python-package")



"""Tests for multi-target training."""

# pylint: disable=unbalanced-tuple-unpacking
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import pytest
from sklearn.datasets import (
    make_classification,
    make_multilabel_classification,
    make_regression,
)
from sklearn.metrics.pairwise import cosine_similarity

import xgboost.testing as tm

from .._typing import ArrayLike
from ..compat import import_cupy
from ..core import Booster, DMatrix, ExtMemQuantileDMatrix, QuantileDMatrix, build_info
from ..objective import Objective, TreeObjective
from ..sklearn import XGBClassifier
from ..training import train
from .data import IteratorForTest
from .updater import ResetStrategy, train_result
from .utils import Device, assert_allclose, non_increasing


def run_multiclass(device: Device, learning_rate: Optional[float]) -> None:
    """Use vector leaf for multi-class models."""
    X, y = make_classification(
        128, n_features=12, n_informative=10, n_classes=4, random_state=2025
    )
    clf = XGBClassifier(
        debug_synchronize=True,
        multi_strategy="multi_output_tree",
        callbacks=[ResetStrategy()],
        n_estimators=10,
        device=device,
        learning_rate=learning_rate,
    )
    clf.fit(X, y, eval_set=[(X, y)])
    assert clf.objective == "multi:softprob"
    assert non_increasing(clf.evals_result()["validation_0"]["mlogloss"])
    if learning_rate is not None and abs(learning_rate - 1.0) < 1e-5:
        assert clf.evals_result()["validation_0"]["mlogloss"][-1] < 0.045

    proba = clf.predict_proba(X)
    assert proba.shape == (y.shape[0], 4)


def run_multilabel(device: Device, learning_rate: Optional[float]) -> None:
    """Use vector leaf for multi-label classification models."""
    X, y = make_multilabel_classification(128, random_state=2025)
    clf = XGBClassifier(
        debug_synchronize=True,
        multi_strategy="multi_output_tree",
        callbacks=[ResetStrategy()],
        n_estimators=10,
        device=device,
        learning_rate=learning_rate,
    )
    clf.fit(X, y, eval_set=[(X, y)])
    assert clf.objective == "binary:logistic"
    assert non_increasing(clf.evals_result()["validation_0"]["logloss"])
    if learning_rate is not None and abs(learning_rate - 1.0) < 1e-5:
        assert clf.evals_result()["validation_0"]["logloss"][-1] < 0.065

    proba = clf.predict_proba(X)
    assert proba.shape == y.shape


def run_quantile_loss(device: Device, weighted: bool) -> None:
    """Check quantile regression for vector leaf."""
    params = {
        "objective": "reg:quantileerror",
        "device": device,
        "quantile_alpha": [0.45, 0.5, 0.55],
        "multi_strategy": "multi_output_tree",
    }
    n_samples = 2048
    X, y = make_regression(n_samples=n_samples, n_features=16, random_state=2026)

    def no_crossing_first_tree(weight: Optional[np.ndarray]) -> None:
        """The first tree should not generate quantile crossing given sufficient amount
        of samples for quantile interpolation.

        """
        Xy = QuantileDMatrix(X, y, weight=weight)
        booster = train(params, Xy, evals=[(Xy, "Train")], num_boost_round=1)
        y_predt = booster.predict(Xy)
        assert y_predt.shape == (n_samples, 3)
        assert (y_predt[:, 0] <= y_predt[:, 1]).all()
        assert (y_predt[:, 1] <= y_predt[:, 2]).all()

    if not weighted:
        weight = None
    else:
        # Test with weights.
        rng = np.random.default_rng(2026)
        weight = rng.uniform(0.0, 1.0, size=n_samples)

    no_crossing_first_tree(weight)

    Xy = QuantileDMatrix(X, y, weight=weight)
    evals_result = train_result(params, Xy, num_rounds=10)
    assert non_increasing(evals_result["train"]["quantile"])


def run_expectile_loss(device: Device) -> None:
    """Check that reparameterized expectile regression produces non-crossing curves."""
    params = {
        "objective": "reg:expectileerror",
        "device": device,
        "expectile_alpha": [0.45, 0.5, 0.55],
        "multi_strategy": "multi_output_tree",
        "tree_method": "hist",
    }
    n_samples = 2048
    rng = np.random.default_rng(2027)
    X = rng.normal(size=(n_samples, 16))
    y = X[:, 0] - 0.5 * X[:, 1] + rng.normal(size=n_samples)

    Xy = QuantileDMatrix(X, y)
    evals_result: Dict[str, Dict] = {}
    booster = train(
        params,
        Xy,
        evals=[(Xy, "Train")],
        evals_result=evals_result,
        num_boost_round=10,
    )

    y_predt = booster.predict(Xy)
    assert y_predt.shape == (n_samples, 3)
    assert (y_predt[:, 0] <= y_predt[:, 1]).all()
    assert (y_predt[:, 1] <= y_predt[:, 2]).all()
    assert non_increasing(evals_result["Train"]["expectile"])


def run_absolute_error(device: Device) -> None:
    """Test mean absolute error with vector leaf."""
    params = {
        "objective": "reg:absoluteerror",
        "device": device,
        "multi_strategy": "multi_output_tree",
    }
    n_samples = 1024
    X, y = make_regression(
        n_samples=n_samples, n_features=16, n_targets=3, random_state=2026
    )
    Xy = QuantileDMatrix(X, y)
    evals_result: Dict[str, Dict] = {}
    train(
        params,
        Xy,
        evals=[(Xy, "Train")],
        verbose_eval=False,
        evals_result=evals_result,
        num_boost_round=16,
    )
    assert non_increasing(evals_result["Train"]["mae"])
    assert evals_result["Train"]["mae"][-1] < 30.0


def _array_impl(device: Device) -> ModuleType:
    if device == "cuda":
        nda = import_cupy()
    else:
        nda = np
    return nda


class LsObj0(TreeObjective):
    """Split grad is the same as value grad."""

    def __init__(self, device: Device) -> None:
        self.device = device

    def __call__(
        self, iteration: int, y_pred: ArrayLike, dtrain: DMatrix
    ) -> Tuple[ArrayLike, ArrayLike]:
        nda = _array_impl(self.device)

        y_true = dtrain.get_label()
        grad, hess = tm.ls_obj(y_true, y_pred, None)
        return nda.array(grad), nda.array(hess)

    def split_grad(
        self, iteration: int, grad: ArrayLike, hess: ArrayLike
    ) -> Tuple[ArrayLike, ArrayLike]:
        nda = _array_impl(self.device)
        return nda.array(grad), nda.array(hess)


class LsObj1(Objective):
    """No split grad."""

    def __init__(self, device: Device) -> None:
        self.device = device

    def __call__(
        self, iteration: int, y_pred: ArrayLike, dtrain: DMatrix
    ) -> Tuple[ArrayLike, ArrayLike]:
        nda = _array_impl(self.device)

        y_true = dtrain.get_label()
        grad, hess = tm.ls_obj(y_true, y_pred, None)
        return nda.array(grad), nda.array(hess)


# Use mean gradient, should still converge.
class LsObj2(LsObj0):
    """Use mean as split grad."""

    def __init__(self, device: Device, check_used: bool):
        self._chk = check_used
        super().__init__(device=device)

    def split_grad(
        self, iteration: int, grad: ArrayLike, hess: ArrayLike
    ) -> Tuple[np.ndarray, np.ndarray]:
        nda = _array_impl(self.device)

        if self._chk:
            assert False
        sgrad = nda.mean(grad, axis=1)
        shess = nda.mean(hess, axis=1)
        return sgrad, shess


def run_reduced_grad(device: Device) -> None:
    """Basic test for using reduced gradient for tree splits."""
    X, y = make_regression(
        n_samples=1024, n_features=16, random_state=1994, n_targets=5
    )
    Xy = QuantileDMatrix(X, y)

    def run_test(
        obj: Optional[Objective], base_score: Optional[list[float]] = None
    ) -> Booster:
        evals_result: Dict[str, Dict] = {}
        booster = train(
            {
                "debug_synchronize": True,
                "device": device,
                "multi_strategy": "multi_output_tree",
                "learning_rate": 1,
                "base_score": base_score,
            },
            Xy,
            evals=[(Xy, "Train")],
            obj=obj,
            num_boost_round=8,
            evals_result=evals_result,
        )
        assert non_increasing(evals_result["Train"]["rmse"])
        return booster

    booster_0 = run_test(LsObj0(device))
    booster_1 = run_test(LsObj1(device))
    np.testing.assert_allclose(
        booster_0.inplace_predict(X), booster_1.inplace_predict(X)
    )

    booster_2 = run_test(LsObj0(device), [0.5] * y.shape[1])
    booster_3 = run_test(None, [0.5] * y.shape[1])
    np.testing.assert_allclose(
        booster_2.inplace_predict(X), booster_3.inplace_predict(X)
    )

    run_test(LsObj2(device, False))
    with pytest.raises(AssertionError):
        run_test(LsObj2(device, True))


def run_with_iter(device: Device) -> None:  # pylint: disable=too-many-locals
    """Test vector leaf with external memory."""
    nda = _array_impl(device)

    n_batches = 4
    n_rounds = 8
    n_targets = 3
    intercept = [0.5] * n_targets

    params = {
        "device": device,
        "multi_strategy": "multi_output_tree",
        "learning_rate": 1.0,
        "base_score": intercept,
        "debug_synchronize": True,
    }

    Xs = []
    ys = []
    for i in range(n_batches):
        X_i, y_i = make_regression(
            n_samples=4096, n_features=8, random_state=(i + 1), n_targets=n_targets
        )
        Xs.append(nda.asarray(X_i))
        ys.append(nda.asarray(y_i))
    it = IteratorForTest(Xs, ys, None, cache="cache", on_host=True)
    Xy: DMatrix = ExtMemQuantileDMatrix(
        it, cache_host_ratio=1.0 if device == "cuda" else None
    )

    evals_result_0: Dict[str, Dict] = {}
    booster_0 = train(
        params,
        Xy,
        num_boost_round=n_rounds,
        evals=[(Xy, "Train")],
        evals_result=evals_result_0,
    )

    it = IteratorForTest(Xs, ys, None, cache=None)
    Xy = QuantileDMatrix(it)
    evals_result_1: Dict[str, Dict] = {}
    booster_1 = train(
        params,
        Xy,
        num_boost_round=n_rounds,
        evals=[(Xy, "Train")],
        evals_result=evals_result_1,
    )
    np.testing.assert_allclose(
        evals_result_0["Train"]["rmse"], evals_result_1["Train"]["rmse"]
    )
    assert non_increasing(evals_result_0["Train"]["rmse"])
    X, _, _ = it.as_arrays()
    assert_allclose(device, booster_0.inplace_predict(X), booster_1.inplace_predict(X))

    binfo = build_info()
    tv = "THRUST_VERSION"
    if device == "cuda" and tv in binfo and binfo[tv][0] < 3:
        pytest.xfail("CCCL version too old.")

    it = IteratorForTest(
        Xs,
        ys,
        None,
        cache="cache",
        on_host=True,
        min_cache_page_bytes=X.shape[0] // n_batches * X.shape[1],
    )
    Xy = ExtMemQuantileDMatrix(it, cache_host_ratio=1.0 if device == "cuda" else None)

    evals_result_2: Dict[str, Dict] = {}
    booster_2 = train(
        params,
        Xy,
        evals=[(Xy, "Train")],
        obj=LsObj0(device),
        num_boost_round=n_rounds,
        evals_result=evals_result_2,
    )
    np.testing.assert_allclose(
        evals_result_0["Train"]["rmse"], evals_result_2["Train"]["rmse"]
    )
    assert_allclose(device, booster_0.inplace_predict(X), booster_2.inplace_predict(X))


def run_eta(device: Device) -> None:
    """Test for learning rate."""
    X, y = make_regression(512, 16, random_state=2025, n_targets=3)

    def run(obj: Optional[Objective]) -> None:
        params = {
            "device": device,
            "multi_strategy": "multi_output_tree",
            "learning_rate": 1.0,
            "debug_synchronize": True,
            "base_score": 0.0,
        }
        Xy = QuantileDMatrix(X, y)
        booster_0 = train(params, Xy, num_boost_round=1, obj=obj)
        params["learning_rate"] = 0.1
        booster_1 = train(params, Xy, num_boost_round=1, obj=obj)
        params["learning_rate"] = 2.0
        booster_2 = train(params, Xy, num_boost_round=1, obj=obj)

        predt_0 = booster_0.predict(Xy)
        predt_1 = booster_1.predict(Xy)
        predt_2 = booster_2.predict(Xy)

        np.testing.assert_allclose(predt_0, predt_1 * 10, rtol=1e-6)
        np.testing.assert_allclose(predt_0 * 2, predt_2, rtol=1e-6)

    run(None)
    run(LsObj0(device))


def run_deterministic(device: Device) -> None:
    """Check the vector leaf implementation is deterministic."""
    X, y = make_regression(
        n_samples=int(2**16), n_features=64, random_state=1994, n_targets=5
    )

    def run() -> Booster:
        Xy = QuantileDMatrix(X, y)
        params = {
            "device": device,
            "multi_strategy": "multi_output_tree",
            "debug_synchronize": True,
        }
        return train(params, Xy, num_boost_round=16)

    booster_0 = run()
    booster_1 = run()
    raw_0 = booster_0.save_raw()
    raw_1 = booster_1.save_raw()
    assert raw_0 == raw_1


def run_column_sampling(device: Device) -> None:
    """Test column sampling with feature importance for multi-target trees."""
    n_features = 32
    X, y = make_regression(
        n_samples=1024, n_features=n_features, random_state=1994, n_targets=3
    )
    # First half of features have weight, second half has 0 weight (not sampled).
    feature_weights = np.zeros(shape=(n_features, 1), dtype=np.float32)
    feature_weights[: n_features // 2] = 1.0 / (n_features / 2)
    Xy = QuantileDMatrix(X, y, feature_weights=feature_weights)

    params = {
        "device": device,
        "multi_strategy": "multi_output_tree",
        "debug_synchronize": True,
        "colsample_bynode": 0.4,
    }
    booster = train(params, Xy, num_boost_round=16)

    # Test all importance types
    for importance_type in ["weight", "gain", "total_gain", "cover", "total_cover"]:
        scores: dict = booster.get_score(importance_type=importance_type)
        assert len(scores) > 0, f"No scores for {importance_type}"

        # Sampled features (first half) should be in scores
        for f in range(0, n_features // 2):
            assert f"f{f}" in scores, f"f{f} not in {importance_type} scores"

        # Non-sampled features (second half) should NOT be in scores
        for f in range(n_features // 2, n_features):
            assert f"f{f}" not in scores

        for score in scores.values():
            assert isinstance(score, float)
            assert score >= 0

    # sklearn Coef
    X, y = make_multilabel_classification(random_state=1994)
    clf = XGBClassifier(
        multi_strategy="multi_output_tree",
        importance_type="weight",
        device=device,
        colsample_bynode=0.2,
    )
    clf.fit(X, y, feature_weights=np.arange(0, X.shape[1]))
    fi = clf.feature_importances_
    assert fi[0] == 0.0
    assert fi[-1] > fi[1] * 5

    w = np.polynomial.Polynomial.fit(np.arange(0, X.shape[1]), fi, deg=1)
    assert w.coef[1] > 0.03


def run_grow_policy(device: Device, grow_policy: str) -> None:
    """Test grow policy (depthwise and lossguide) for vector leaf."""
    X, y = make_regression(
        n_samples=1024, n_features=16, random_state=1994, n_targets=3
    )
    Xy = QuantileDMatrix(X, y)

    params = {
        "device": device,
        "multi_strategy": "multi_output_tree",
        "debug_synchronize": True,
        "grow_policy": grow_policy,
    }

    evals_result = train_result(params, Xy, num_rounds=10)
    assert non_increasing(evals_result["train"]["rmse"])


def run_mixed_strategy(device: Device) -> None:
    """Test mixed multi_strategy with ResetStrategy callback."""
    X, y = make_classification(
        n_samples=1024, n_informative=8, n_classes=3, random_state=1994
    )
    Xy = DMatrix(data=X, label=y)

    booster = train(
        {
            "num_parallel_tree": 4,
            "num_class": 3,
            "objective": "multi:softprob",
            "multi_strategy": "multi_output_tree",
            "device": device,
            "debug_synchronize": True,
            "base_score": 0,
        },
        num_boost_round=16,
        dtrain=Xy,
        callbacks=[ResetStrategy()],
    )

    # Test model slicing - each boosting round should be iterable
    assert len(list(booster)) == 16

    # Test that sliced predictions sum to full prediction
    predt = booster.predict(Xy, output_margin=True)
    predt_sum = np.zeros(predt.shape)
    for t in booster:
        predt_sum += t.predict(Xy, output_margin=True)
    np.testing.assert_allclose(predt, predt_sum, atol=1e-5)

    # Test feature importance works with mixed trees
    for importance_type in ["weight", "gain", "total_gain", "cover", "total_cover"]:
        scores = booster.get_score(importance_type=importance_type)
        assert len(scores) > 0
        for score in scores.values():
            assert isinstance(score, float)
            assert score >= 0


def run_feature_importance_strategy_compare(device: Device) -> None:
    """Different strategies produce similar feature importance ratios."""
    n_features = 16
    X, y = make_classification(
        n_samples=2048,
        n_features=n_features,
        n_informative=10,
        n_classes=4,
        random_state=1994,
    )
    Xy = DMatrix(data=X, label=y)

    base_params: Dict[str, Any] = {
        "num_class": 4,
        "objective": "multi:softprob",
        "device": device,
        "debug_synchronize": True,
        "max_depth": 5,
    }

    # Train models with different strategies
    boosters = [
        train(
            {**base_params, "multi_strategy": "multi_output_tree"},
            Xy,
            num_boost_round=32,
        ),
        train(
            {**base_params, "multi_strategy": "one_output_per_tree"},
            Xy,
            num_boost_round=32,
        ),
        train(
            {**base_params, "multi_strategy": "multi_output_tree"},
            Xy,
            num_boost_round=32,
            callbacks=[ResetStrategy()],
        ),
    ]

    def get_normalized_importance(booster: Booster, importance_type: str) -> np.ndarray:
        """Get feature importance as normalized array (sums to 1)."""
        scores = booster.get_score(importance_type=importance_type)
        arr = np.array([scores.get(f"f{i}", 0.0) for i in range(n_features)])
        return arr / arr.sum() if arr.sum() > 0 else arr

    for importance_type in ["weight", "gain", "total_gain", "cover", "total_cover"]:
        imps = [get_normalized_importance(b, importance_type) for b in boosters]

        # Check that importances are not exactly the same (different strategies)
        assert not np.allclose(imps[0], imps[1])
        assert not np.allclose(imps[0], imps[2])

        # Check that normalized importances are similar (correlated)
        # All strategies should have reasonably similar importance patterns
        assert cosine_similarity([imps[0]], [imps[1]])[0, 0] > 0.9
        assert cosine_similarity([imps[0]], [imps[2]])[0, 0] > 0.9
        assert cosine_similarity([imps[1]], [imps[2]])[0, 0] > 0.9


# pylint: disable=too-many-arguments, too-many-locals
def _run_regression_objective_test(
    device: Device,
    objective: str,
    metric: str,
    X: np.ndarray,
    y: np.ndarray,
    *,
    extra_params: Optional[Dict[str, Any]] = None,
    check_pred_positive: bool = False,
    check_pred_probability: bool = False,
    check_pred_binary: bool = False,
    strictly_non_increasing: bool = True,
) -> None:
    params: Dict[str, Any] = {
        "objective": objective,
        "device": device,
        "multi_strategy": "multi_output_tree",
    }
    if extra_params:
        params.update(extra_params)

    n_samples = X.shape[0]
    n_targets = y.shape[1] if y.ndim > 1 else 1

    Xy = DMatrix(X, y)
    evals_result: Dict[str, Dict] = {}
    booster = train(
        params,
        Xy,
        evals=[(Xy, "Train")],
        verbose_eval=False,
        evals_result=evals_result,
        num_boost_round=16,
    )
    predt = booster.predict(Xy)
    assert predt.shape == (n_samples, n_targets)

    if check_pred_positive:
        assert (predt > 0).all()
    if check_pred_probability:
        assert (predt > 0).all() and (predt < 1).all()
    if check_pred_binary:
        assert set(np.unique(predt)).issubset({0.0, 1.0})

    metric_vals = evals_result["Train"][metric]
    if strictly_non_increasing:
        assert non_increasing(metric_vals)
    else:
        assert metric_vals[-1] < metric_vals[0]


def run_reg_squarederror(device: Device) -> None:
    """Test squared error regression with vector leaf."""
    n_samples, n_targets = 1024, 3
    X, y = make_regression(
        n_samples=n_samples, n_features=16, n_targets=n_targets, random_state=2026
    )
    _run_regression_objective_test(device, "reg:squarederror", "rmse", X, y)


def run_reg_logistic(device: Device) -> None:
    """Test logistic regression for probability with vector leaf."""
    n_samples, n_targets = 1024, 3
    rng = np.random.default_rng(2026)
    X = rng.standard_normal((n_samples, 16))
    y = rng.uniform(0.0, 1.0, (n_samples, n_targets))  # Labels in [0, 1]
    _run_regression_objective_test(
        device, "reg:logistic", "rmse", X, y, check_pred_probability=True
    )


def run_reg_gamma(device: Device) -> None:
    """Test gamma regression with vector leaf."""
    n_samples, n_targets = 1024, 3
    rng = np.random.default_rng(2026)
    X = rng.standard_normal((n_samples, 16))
    y = rng.gamma(2.0, 2.0, (n_samples, n_targets))  # Labels must be positive
    _run_regression_objective_test(
        device, "reg:gamma", "gamma-deviance", X, y, check_pred_positive=True
    )


def run_reg_squaredlogerror(device: Device) -> None:
    """Test squared log error regression with vector leaf."""
    n_samples, n_targets = 1024, 3
    rng = np.random.default_rng(2026)
    X = rng.standard_normal((n_samples, 16))
    y = np.abs(rng.standard_normal((n_samples, n_targets))) + 0.1  # Labels > -1
    _run_regression_objective_test(device, "reg:squaredlogerror", "rmsle", X, y)


def run_reg_pseudohubererror(device: Device) -> None:
    """Test pseudo huber error regression with vector leaf."""
    n_samples, n_targets = 1024, 3
    X, y = make_regression(
        n_samples=n_samples, n_features=16, n_targets=n_targets, random_state=2026
    )
    _run_regression_objective_test(
        device, "reg:pseudohubererror", "mphe", X, y, extra_params={"huber_slope": 1.0}
    )


def run_binary_logitraw(device: Device) -> None:
    """Test binary logitraw with vector leaf (multi-label classification)."""
    n_samples = 1024
    X, y = make_multilabel_classification(n_samples, random_state=2026)
    _run_regression_objective_test(
        device, "binary:logitraw", "logloss", X, y, strictly_non_increasing=False
    )


def run_binary_hinge(device: Device) -> None:
    """Test binary hinge loss with vector leaf (multi-label classification)."""
    n_samples = 1024
    X, y = make_multilabel_classification(n_samples, random_state=2026)
    _run_regression_objective_test(
        device,
        "binary:hinge",
        "error",
        X,
        y,
        check_pred_binary=True,
        strictly_non_increasing=False,
    )


def run_count_poisson(device: Device) -> None:
    """Test Poisson regression with vector leaf."""
    n_samples, n_targets = 1024, 3
    rng = np.random.default_rng(2026)
    X = rng.standard_normal((n_samples, 16))
    y = rng.poisson(5, (n_samples, n_targets)).astype(np.float32)  # Labels >= 0
    _run_regression_objective_test(
        device, "count:poisson", "poisson-nloglik", X, y, check_pred_positive=True
    )


def run_reg_tweedie(device: Device) -> None:
    """Test Tweedie regression with vector leaf."""
    n_samples, n_targets = 1024, 3
    rng = np.random.default_rng(2026)
    X = rng.standard_normal((n_samples, 16))
    y = rng.gamma(2.0, 2.0, (n_samples, n_targets))  # Labels >= 0
    _run_regression_objective_test(
        device, "reg:tweedie", "tweedie-nloglik@1.5", X, y, check_pred_positive=True
    )


def all_reg_objectives() -> List[Callable[[Device], None]]:
    """List of obj tests."""
    objs: List[Callable[[Device], None]] = [
        run_reg_squarederror,
        run_reg_logistic,
        run_reg_gamma,
        run_reg_squaredlogerror,
        run_reg_pseudohubererror,
        run_binary_logitraw,
        run_binary_hinge,
        run_count_poisson,
        run_reg_tweedie,
    ]
    return objs


def _make_subsample_params(device: Device, sampling_method: str) -> dict:
    params = {
        "device": device,
        "tree_method": "hist",
        "multi_strategy": "multi_output_tree",
        "subsample": 0.5,
        "sampling_method": sampling_method,
        "max_depth": 6,
        "debug_synchronize": True,
        "seed": 2026,
    }
    return params


def run_subsample(device: Device, sampling_method: str) -> None:
    """Test row subsampling."""
    n_samples = 2048
    X, y = make_regression(
        n_samples=n_samples, n_features=16, n_targets=3, random_state=2026
    )
    Xy = QuantileDMatrix(X, y)

    params = _make_subsample_params(device, sampling_method)

    evals_result = train_result(params, Xy, num_rounds=16)
    # Training should converge with subsampling
    assert non_increasing(evals_result["train"]["rmse"], tolerance=0.01)

    # Test with quantile regression
    params = _make_subsample_params(device, sampling_method)
    params["objective"] = "reg:quantileerror"
    params["quantile_alpha"] = [0.25, 0.5, 0.75]
    Xy_single = QuantileDMatrix(X, y[:, 0])
    evals_result_q = train_result(params, Xy_single, num_rounds=16)
    assert non_increasing(evals_result_q["train"]["quantile"], tolerance=0.01)


def run_gradient_based_sampling_accuracy(device: Device) -> None:
    """Test that gradient-based sampling provides better accuracy."""
    n_samples = 4096
    X, y = make_regression(
        n_samples=n_samples, n_features=16, n_targets=3, random_state=2026
    )
    Xy = QuantileDMatrix(X, y)

    params_uniform = _make_subsample_params(device, "uniform")

    def run(obj: Callable | None) -> None:
        # Train with uniform sampling
        evals_uniform: Dict[str, Dict] = {}
        train(
            params_uniform,
            Xy,
            num_boost_round=32,
            evals=[(Xy, "train")],
            obj=obj,
            verbose_eval=False,
            evals_result=evals_uniform,
        )

        # Train with gradient-based sampling
        params_grad = _make_subsample_params(device, "gradient_based")
        evals_grad: Dict[str, Dict] = {}
        train(
            params_grad,
            Xy,
            num_boost_round=32,
            evals=[(Xy, "train")],
            obj=obj,
            verbose_eval=False,
            evals_result=evals_grad,
        )

        uniform_final = evals_uniform["train"]["rmse"][-1]
        grad_final = evals_grad["train"]["rmse"][-1]
        assert non_increasing(evals_uniform["train"]["rmse"])
        assert non_increasing(evals_grad["train"]["rmse"])
        if obj is None:
            assert grad_final < uniform_final

    run(None)
    run(LsObj2(device, False))



"""
A demo for multi-output regression using reduced gradient
=========================================================

See :doc:`/tutorials/multioutput` for more information.

.. versionadded:: 3.2.0

.. note::

    The implementation is experimental and many features are missing.

.. seealso:: :ref:`sphx_glr_python_examples_multioutput_regression.py`

"""

import argparse
from typing import Tuple

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.datasets import make_regression

import xgboost as xgb
from xgboost.objective import TreeObjective


class LsObjMean(TreeObjective):
    """Least squared error. Reduce the size of the gradient using mean value."""

    def __init__(self, device: str) -> None:
        self.device = device

    def __call__(
        self, iteration: int, y_pred: np.ndarray, dtrain: xgb.DMatrix
    ) -> Tuple[np.ndarray, np.ndarray]:
        y_true = dtrain.get_label()
        grad = y_pred - y_true
        if self.device == "cpu":
            hess = np.ones(grad.shape)
            return grad, hess
        else:
            import cupy as cp

            hess = cp.ones(grad.shape)

            return cp.array(grad), cp.array(hess)

    def split_grad(
        self, iteration: int, grad: np.ndarray, hess: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        if self.device == "cpu":
            from numpy import mean
        else:
            from cupy import mean  # type: ignore[no-redef]

        sgrad = mean(grad, axis=1)
        shess = mean(hess, axis=1)
        return sgrad, shess


def svd_class(device: str) -> BaseEstimator:
    """One of the methods in the sketch boost paper."""
    from sklearn.decomposition import TruncatedSVD

    svd_params = {"algorithm": "arpack", "n_components": 2, "n_iter": 8}
    svd = TruncatedSVD(**svd_params)
    return svd


class LsObjSvd(LsObjMean):
    """Reduce the size of the gradient using SVD."""

    def __init__(self, device: str) -> None:
        super().__init__(device=device)

    def split_grad(
        self, iteration: int, grad: np.ndarray, hess: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        svd = svd_class(self.device)
        if self.device == "cuda":
            grad = grad.get()   # type: ignore
            hess = hess.get()   # type: ignore

        svd.fit(grad)
        grad = svd.transform(grad)
        hess = svd.transform(hess)
        if self.device == "cpu":
            hess = np.clip(hess, 0.01, None)
        else:
            import cupy as cp

            hess = cp.clip(hess, 0.01, None)
        return grad, hess


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--device", choices=["cpu", "cuda"], default="cpu")
    args = parser.parse_args()

    X, y = make_regression(
        n_samples=8192, n_features=32, n_targets=8, random_state=2026
    )
    Xy = xgb.QuantileDMatrix(X, y)

    for obj in (LsObjMean(args.device), LsObjSvd(args.device)):
        xgb.train(
            {
                "device": args.device,
                "multi_strategy": "multi_output_tree",
            },
            Xy,
            evals=[(Xy, "Train")],
            obj=obj,
            num_boost_round=16,
        )


if __name__ == "__main__":
    main()



# pylint: disable=too-many-arguments, too-many-positional-arguments
"""Tests for compatiblity with sklearn."""

from typing import Callable, Optional, Type

import numpy as np
import pytest
from scipy.special import softmax

from ..core import DMatrix
from ..sklearn import XGBClassifier, XGBRegressor, XGBRFRegressor
from .data import get_california_housing, make_batches
from .ordinal import make_recoded
from .utils import Device, assert_allclose


def run_boost_from_prediction_binary(
    tree_method: str,
    device: Device,
    X: np.ndarray,
    y: np.ndarray,
    as_frame: Optional[Callable],
) -> None:
    """
    Parameters
    ----------

    as_frame: A callable function to convert margin into DataFrame, useful for different
    df implementations.
    """

    model_0 = XGBClassifier(
        learning_rate=0.3,
        random_state=0,
        n_estimators=4,
        tree_method=tree_method,
        device=device,
    )
    model_0.fit(X=X, y=y)
    margin = model_0.predict(X, output_margin=True)
    if as_frame is not None:
        margin = as_frame(margin)

    model_1 = XGBClassifier(
        learning_rate=0.3,
        random_state=0,
        n_estimators=4,
        tree_method=tree_method,
        device=device,
    )
    model_1.fit(X=X, y=y, base_margin=margin)
    predictions_1 = model_1.predict(X, base_margin=margin)

    cls_2 = XGBClassifier(
        learning_rate=0.3,
        random_state=0,
        n_estimators=8,
        tree_method=tree_method,
        device=device,
    )
    cls_2.fit(X=X, y=y)
    predictions_2 = cls_2.predict(X)
    np.testing.assert_allclose(predictions_1, predictions_2)


def run_boost_from_prediction_multi_clasas(
    estimator: Type,
    tree_method: str,
    device: Device,
    X: np.ndarray,
    y: np.ndarray,
    as_frame: Optional[Callable],
) -> None:
    """Boosting from prediction with multi-class clf."""
    # Multi-class
    model_0 = estimator(
        learning_rate=0.3,
        random_state=0,
        n_estimators=4,
        tree_method=tree_method,
        device=device,
    )
    model_0.fit(X=X, y=y)
    margin = model_0.get_booster().inplace_predict(X, predict_type="margin")
    if as_frame is not None:
        margin = as_frame(margin)

    model_1 = estimator(
        learning_rate=0.3,
        random_state=0,
        n_estimators=4,
        tree_method=tree_method,
        device=device,
    )
    model_1.fit(X=X, y=y, base_margin=margin)
    predictions_1 = model_1.get_booster().predict(
        DMatrix(X, base_margin=margin), output_margin=True
    )

    model_2 = estimator(
        learning_rate=0.3,
        random_state=0,
        n_estimators=8,
        tree_method=tree_method,
        device=device,
    )
    model_2.fit(X=X, y=y)
    predictions_2 = model_2.get_booster().inplace_predict(X, predict_type="margin")

    if hasattr(predictions_1, "get"):
        predictions_1 = predictions_1.get()
    if hasattr(predictions_2, "get"):
        predictions_2 = predictions_2.get()
    np.testing.assert_allclose(predictions_1, predictions_2, atol=1e-6)


def run_housing_rf_regression(tree_method: str, device: Device) -> None:
    """Testwith the cali housing dataset."""
    from sklearn.metrics import mean_squared_error
    from sklearn.model_selection import KFold

    X, y = get_california_housing()
    rng = np.random.RandomState(1994)
    kf = KFold(n_splits=2, shuffle=True, random_state=rng)
    for train_index, test_index in kf.split(X, y):
        xgb_model = XGBRFRegressor(
            random_state=42, tree_method=tree_method, device=device
        ).fit(X[train_index], y[train_index])
        preds = xgb_model.predict(X[test_index])
        labels = y[test_index]
        assert mean_squared_error(preds, labels) < 35

    rfreg = XGBRFRegressor(device=device)
    with pytest.raises(NotImplementedError):
        rfreg.set_params(early_stopping_rounds=10)
        rfreg.fit(X, y)


def run_recoding(device: Device) -> None:
    """Test re-coding for training continuation."""
    enc, reenc, y, _, _ = make_recoded(device, n_features=16)
    reg = XGBRegressor(enable_categorical=True, n_estimators=2, device=device)
    reg.fit(enc, y, eval_set=[(reenc, y)])
    results_0 = reg.evals_result()

    booster = reg.get_booster()
    assert not booster.get_categories().empty()

    reg = XGBRegressor(enable_categorical=True, n_estimators=2, device=device)
    reg.fit(reenc, y, xgb_model=booster, eval_set=[(enc, y)])
    results_1 = reg.evals_result()

    booster = reg.get_booster()
    assert booster.num_boosted_rounds() == 4
    assert not booster.get_categories().empty()

    reg = XGBRegressor(enable_categorical=True, n_estimators=4, device=device)
    reg.fit(enc, y, eval_set=[(reenc, y)])
    results_2 = reg.evals_result()

    np.testing.assert_allclose(
        results_2["validation_0"]["rmse"],
        results_0["validation_0"]["rmse"] + results_1["validation_0"]["rmse"],
    )

    np.testing.assert_allclose(reg.predict(reenc), reg.predict(enc))
    np.testing.assert_allclose(reg.apply(reenc), reg.apply(enc))


def run_intercept(device: Device) -> None:
    """Tests for the intercept."""
    from sklearn.datasets import make_classification, make_multilabel_classification

    X, y, w = [v[0] for v in make_batches(256, 3, 1, use_cupy=False)]
    reg = XGBRegressor(device=device)
    reg.fit(X, y, sample_weight=w)
    result = reg.intercept_
    assert result.dtype == np.float32
    assert result[0] < 0.5

    reg = XGBRegressor(booster="gblinear", device=device)
    reg.fit(X, y, sample_weight=w)
    result = reg.intercept_
    assert isinstance(result, np.ndarray)
    assert result.dtype == np.float32
    assert result[0] < 0.5

    n_classes = 4
    X, y = make_classification(
        random_state=1994,
        n_samples=128,
        n_features=16,
        n_classes=n_classes,
        n_informative=16,
        n_redundant=0,
    )

    clf = XGBClassifier(booster="gbtree", objective="multi:softprob", device=device)
    clf.fit(X, y)
    result = clf.intercept_
    assert isinstance(result, np.ndarray)
    assert len(result) == 4

    assert (softmax(result) >= 0.0).all()
    np.testing.assert_allclose(sum(result), 0.0, atol=1e-6)
    np.testing.assert_allclose(sum(softmax(result)), 1.0)

    # Tests for user input
    # Multi-class
    intercept = np.ones(shape=(n_classes), dtype=np.float32) / n_classes
    if device == "cuda":
        import cupy as cp

        intercept = cp.array(intercept)

    clf = XGBClassifier(objective="multi:softprob", base_score=intercept)
    clf.fit(X, y)
    assert_allclose(device, intercept, clf.intercept_)

    X, y = make_multilabel_classification(  # pylint: disable=unbalanced-tuple-unpacking
        random_state=1994, n_samples=128, n_features=16, n_classes=n_classes
    )

    # Multi-label
    intercept = np.ones(shape=(n_classes), dtype=np.float32) / 2
    if device == "cuda":
        import cupy as cp

        intercept = cp.array(intercept)

    clf = XGBClassifier(base_score=intercept)
    clf.fit(X, y)
    assert_allclose(device, intercept, clf.intercept_)
    assert clf.objective == "binary:logistic"



"""
Demo for obtaining leaf index
=============================
"""
import os

import xgboost as xgb

# load data in do training
CURRENT_DIR = os.path.dirname(__file__)
dtrain = xgb.DMatrix(
    os.path.join(CURRENT_DIR, "../data/agaricus.txt.train?format=libsvm")
)
dtest = xgb.DMatrix(
    os.path.join(CURRENT_DIR, "../data/agaricus.txt.test?format=libsvm")
)
param = {"max_depth": 2, "eta": 1, "objective": "binary:logistic"}
watchlist = [(dtest, "eval"), (dtrain, "train")]
num_round = 3
bst = xgb.train(param, dtrain, num_round, watchlist)

print("start testing predict the leaf indices")
# predict using first 2 tree
leafindex = bst.predict(
    dtest, iteration_range=(0, 2), pred_leaf=True, strict_shape=True
)
print(leafindex.shape)
print(leafindex)
# predict all trees
leafindex = bst.predict(dtest, pred_leaf=True)
print(leafindex.shape)



import numpy as np
import pytest
import xgboost as xgb
from xgboost import testing as tm
from xgboost.testing.parse_tree import (
    run_split_value_histograms,
    run_tree_to_df_categorical,
)

pytestmark = pytest.mark.skipif(**tm.no_pandas())


dpath = "demo/data/"
rng = np.random.RandomState(1994)


class TestTreesToDataFrame:
    def build_model(self, max_depth, num_round):
        dtrain, _ = tm.load_agaricus(__file__)
        param = {"max_depth": max_depth, "objective": "binary:logistic", "verbosity": 1}
        num_round = num_round
        bst = xgb.train(param, dtrain, num_round)
        return bst

    def parse_dumped_model(self, booster, item_to_get, splitter):
        item_to_get += "="
        txt_dump = booster.get_dump(with_stats=True)
        tree_list = [tree.split("/n") for tree in txt_dump]
        split_trees = [tree[0].split(item_to_get)[1:] for tree in tree_list]
        res = sum(
            [float(line.split(splitter)[0]) for tree in split_trees for line in tree]
        )
        return res

    def test_trees_to_dataframe(self):
        bst = self.build_model(max_depth=5, num_round=10)
        gain_from_dump = self.parse_dumped_model(
            booster=bst, item_to_get="gain", splitter=","
        )
        cover_from_dump = self.parse_dumped_model(
            booster=bst, item_to_get="cover", splitter="\n"
        )
        # method being tested
        df = bst.trees_to_dataframe()

        # test for equality of gains
        gain_from_df = df[df.Feature != "Leaf"][["Gain"]].sum()
        assert np.allclose(gain_from_dump, gain_from_df)

        # test for equality of covers
        cover_from_df = df.Cover.sum()
        assert np.allclose(cover_from_dump, cover_from_df)

    def test_tree_to_df_categorical(self) -> None:
        run_tree_to_df_categorical("approx", "cpu")

    def test_tree_to_df_indicator(self, tmp_path) -> None:
        """Test trees_to_dataframe with indicator (boolean) features."""
        n_samples = 200
        n_features = 5
        X_int = rng.randint(0, 2, size=(n_samples, n_features))
        y = np.logical_xor(X_int[:, 0], X_int[:, 1]).astype(np.float32)
        X = X_int.astype(np.float32)
        dtrain = xgb.DMatrix(X, label=y)

        # Create a feature map with indicator type 'i'
        fmap_path = str(tmp_path / "fmap.txt")
        with open(fmap_path, "w", encoding="utf-8") as f:
            for i in range(n_features):
                f.write(f"{i}\tf{i}\ti\n")

        bst = xgb.train(
            {"max_depth": 3, "objective": "binary:logistic", "verbosity": 0},
            dtrain,
            num_boost_round=5,
        )
        df = bst.trees_to_dataframe(fmap=fmap_path)

        # Basic structure checks
        assert "Tree" in df.columns
        assert "Feature" in df.columns
        assert "Gain" in df.columns
        assert "Cover" in df.columns
        assert len(df) > 0

        # Indicator nodes should have NaN splits; missing defaults to no-direction
        non_leaf = df[df.Feature != "Leaf"]
        assert len(non_leaf) > 0
        assert non_leaf["Split"].isna().all()
        assert (non_leaf["Missing"] == non_leaf["No"]).all()

    def test_split_value_histograms(self):
        run_split_value_histograms("approx", "cpu")



"""Create Package variants for PyPI distribution."""

import argparse
import os
import tomllib

from packaging.version import Version
from test_utils import PY_PACKAGE

IN_PATH = os.path.join(PY_PACKAGE, "pyproject.toml.in")
STUB_IN_PATH = os.path.join(PY_PACKAGE, "pyproject.toml.stub.in")
OUT_PATH = os.path.join(PY_PACKAGE, "pyproject.toml")

NCCL_WHL = """    \"nvidia-nccl-{0} ; platform_system == 'Linux'\","""

NAME = "{{ name }}"
NCCL = "{{ nccl }}"
VERSION = "{{ version }}"
CUDA_VARIANTS = ["cu12", "cu13"]


def copyfile(src: str, dst: str) -> None:
    with open(src, "rb") as fd:
        content = fd.read()
    with open(dst, "wb") as fd:
        fd.write(content)


def make_pyproject(
    *, use_suffix: str, require_nccl_dep: str, create_stub: bool = False
) -> None:
    if use_suffix == "cpu" and require_nccl_dep != "na":
        raise ValueError(
            "xgboost-cpu cannot require NCCL dependency. "
            "When setting --use-suffix='cpu', you must also set --require-nccl-dep='na'."
        )
    if (
        use_suffix in CUDA_VARIANTS
        and require_nccl_dep in CUDA_VARIANTS
        and use_suffix != require_nccl_dep
    ):
        raise ValueError(
            "Inconsistent choices for --use-suffix and --require-nccl-dep. "
            "When --use-suffix is set to one of {{{0}}}, --require-nccl-dep must be "
            "set to identical value as --use-suffix.".format(",".join(CUDA_VARIANTS))
        )
    if create_stub:
        if use_suffix == "na":
            raise ValueError("To create a stub package, --use-suffix must not be 'na'")
        if require_nccl_dep != "na":
            raise ValueError(
                "To create a stub package, --require-nccl-dep must be 'na'"
            )

    with open(STUB_IN_PATH if create_stub else IN_PATH) as fd:
        pyproject = fd.read()

    readme_dft = os.path.join(PY_PACKAGE, "README.dft.rst")
    readme_cpu = os.path.join(PY_PACKAGE, "README.cpu.rst")
    readme_stub = os.path.join(PY_PACKAGE, "README.stub.rst")
    readme = os.path.join(PY_PACKAGE, "README.rst")
    pyproject = pyproject.replace(
        NAME, f"xgboost-{use_suffix}" if use_suffix != "na" else "xgboost"
    )
    if create_stub:
        copyfile(readme_stub, readme)
        pyproject_parsed = tomllib.loads(pyproject)
        pyproject = pyproject.replace(
            VERSION, str(Version(pyproject_parsed["project"]["version"]))
        )
    elif use_suffix == "cpu":
        copyfile(readme_cpu, readme)
    else:
        copyfile(readme_dft, readme)
    pyproject = pyproject.replace(
        NCCL, NCCL_WHL.format(require_nccl_dep) if require_nccl_dep != "na" else ""
    )
    pyproject = (
        f"# Generated by `{os.path.basename(__file__)}`, don't edit: "
        f"'--use-suffix={use_suffix} --require-nccl-dep={require_nccl_dep} "
        f"--create-stub={create_stub}'\n" + pyproject
    )

    with open(OUT_PATH, "w") as fd:
        fd.write(pyproject)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--use-suffix",
        type=str,
        choices=["na", "cpu"] + CUDA_VARIANTS,
        default="na",
        help=(
            "When using this option, rename the package name to xgboost-[suffix]. "
            "Set to 'na' to disable"
        ),
    )
    parser.add_argument(
        "--require-nccl-dep",
        type=str,
        choices=["na"] + CUDA_VARIANTS,
        required=True,
        help="Which NCCL dependency to use; select 'na' to remove NCCL dependency",
    )
    parser.add_argument(
        "--create-stub",
        action="store_true",
        help="Create a stub package that redirects users to install `xgboost`",
    )
    args = parser.parse_args()
    make_pyproject(
        use_suffix=args.use_suffix,
        require_nccl_dep=args.require_nccl_dep,
        create_stub=args.create_stub,
    )



import os

import numpy as np
import xgboost
from sklearn.datasets import make_classification
from xgboost.testing import make_categorical, make_ltr

kRounds = 4
kRows = 1000
kCols = 4
kForests = 2
kMaxDepth = 3
kClasses = 3


version = xgboost.__version__

target_dir = "models"


def booster_ubj(model: str) -> str:
    return os.path.join(target_dir, "xgboost-" + version + "." + model + ".ubj")


def booster_json(model: str) -> str:
    return os.path.join(target_dir, "xgboost-" + version + "." + model + ".json")


def skl_ubj(model: str) -> str:
    return os.path.join(target_dir, "xgboost_scikit-" + version + "." + model + ".ubj")


def skl_json(model: str) -> str:
    return os.path.join(target_dir, "xgboost_scikit-" + version + "." + model + ".json")


def generate_regression_model() -> None:
    print("Regression")
    X, y = make_categorical(
        n_samples=kRows, n_features=kCols, n_categories=16, onehot=False, cat_ratio=0.5
    )
    w = np.random.default_rng(2025).uniform(size=X.shape[0])
    data = xgboost.DMatrix(X, label=y, weight=w)
    booster = xgboost.train(
        {
            "tree_method": "hist",
            "num_parallel_tree": kForests,
            "max_depth": kMaxDepth,
            "base_score": 0.5,
        },
        num_boost_round=kRounds,
        dtrain=data,
    )
    booster.save_model(booster_ubj("reg"))
    booster.save_model(booster_json("reg"))

    reg = xgboost.XGBRegressor(
        tree_method="hist",
        num_parallel_tree=kForests,
        max_depth=kMaxDepth,
        n_estimators=kRounds,
        base_score=0.5,
    )
    reg.fit(X, y, sample_weight=w)
    reg.save_model(skl_ubj("reg"))
    reg.save_model(skl_json("reg"))


def generate_logistic_model() -> None:
    print("Logistic")
    X, y = make_classification(n_samples=kRows, n_features=kCols, random_state=2025)
    assert y.max() == 1 and y.min() == 0
    w = np.random.default_rng(2025).uniform(size=X.shape[0])

    for objective, name in [
        ("binary:logistic", "logit"),
        ("binary:logitraw", "logitraw"),
    ]:
        data = xgboost.DMatrix(X, label=y, weight=w)
        booster = xgboost.train(
            {
                "tree_method": "hist",
                "num_parallel_tree": kForests,
                "max_depth": kMaxDepth,
                "objective": objective,
                "base_score": 0.5,
            },
            num_boost_round=kRounds,
            dtrain=data,
        )
        booster.save_model(booster_ubj(name))
        booster.save_model(booster_json(name))

        reg = xgboost.XGBClassifier(
            tree_method="hist",
            num_parallel_tree=kForests,
            max_depth=kMaxDepth,
            n_estimators=kRounds,
            objective=objective,
            base_score=0.5,
        )
        reg.fit(X, y, sample_weight=w)
        reg.save_model(skl_ubj(name))
        reg.save_model(skl_json(name))


def generate_classification_model() -> None:
    print("Classification")
    X, y = make_classification(
        n_samples=kRows,
        n_features=kCols,
        random_state=2025,
        n_classes=kClasses,
        n_informative=4,
        n_redundant=0,
    )
    w = np.random.default_rng(2025).uniform(size=X.shape[0])

    data = xgboost.DMatrix(X, label=y, weight=w)
    booster = xgboost.train(
        {
            "num_class": kClasses,
            "tree_method": "hist",
            "num_parallel_tree": kForests,
            "max_depth": kMaxDepth,
        },
        num_boost_round=kRounds,
        dtrain=data,
    )
    booster.save_model(booster_ubj("cls"))
    booster.save_model(booster_json("cls"))

    cls = xgboost.XGBClassifier(
        tree_method="hist",
        num_parallel_tree=kForests,
        max_depth=kMaxDepth,
        n_estimators=kRounds,
    )
    cls.fit(X, y, sample_weight=w)
    cls.save_model(skl_ubj("cls"))
    cls.save_model(skl_json("cls"))


def generate_ranking_model() -> None:
    print("Learning to Rank")
    X, y, qid, w = make_ltr(
        n_samples=kRows, n_features=kCols, n_query_groups=7, max_rel=3
    )

    data = xgboost.DMatrix(X, y, weight=w, qid=qid)
    booster = xgboost.train(
        {
            "objective": "rank:ndcg",
            "num_parallel_tree": kForests,
            "tree_method": "hist",
            "max_depth": kMaxDepth,
            "base_score": 0.5,
        },
        num_boost_round=kRounds,
        dtrain=data,
    )
    booster.save_model(booster_ubj("ltr"))
    booster.save_model(booster_json("ltr"))

    ranker = xgboost.sklearn.XGBRanker(
        n_estimators=kRounds,
        tree_method="hist",
        objective="rank:ndcg",
        max_depth=kMaxDepth,
        num_parallel_tree=kForests,
        base_score=0.5,
    )
    ranker.fit(X, y, qid=qid, sample_weight=w)
    ranker.save_model(skl_ubj("ltr"))
    ranker.save_model(skl_json("ltr"))


def generate_aft_survival_models() -> None:
    print("AFT Survival")
    X, y_lower = make_categorical(
        n_samples=kRows, n_features=kCols, n_categories=16, onehot=False, cat_ratio=0.5
    )
    w = np.random.default_rng(2025).uniform(size=X.shape[0])
    y_upper = y_lower + np.mean(y_lower) + w
    data = xgboost.QuantileDMatrix(
        X, label_lower_bound=y_lower, label_upper_bound=y_upper
    )
    params = {
        "num_parallel_tree": kForests,
        "tree_method": "hist",
        "max_depth": kMaxDepth,
        "objective": "survival:aft",
        "aft_loss_distribution": "normal",
        "base_score": 0.5,
    }
    booster = xgboost.train(params, num_boost_round=kRounds, dtrain=data)
    booster.save_model(booster_ubj("aft"))
    booster.save_model(booster_json("aft"))


if __name__ == "__main__":
    if not os.path.exists(target_dir):
        os.mkdir(target_dir)

    generate_regression_model()
    generate_logistic_model()
    generate_classification_model()
    generate_ranking_model()
    generate_aft_survival_models()



import sys

from sklearn.datasets import make_classification
from sklearn.metrics import roc_auc_score

import xgboost as xgb


def run_omp(output_path: str):
    X, y = make_classification(
        n_samples=200, n_features=32, n_classes=3, n_informative=8
    )
    Xy = xgb.DMatrix(X, y, nthread=16)
    booster = xgb.train(
        {"num_class": 3, "objective": "multi:softprob", "n_jobs": 16},
        Xy,
        num_boost_round=8,
    )
    score = booster.predict(Xy)
    auc = roc_auc_score(y, score, average="weighted", multi_class="ovr")
    with open(output_path, "w") as fd:
        fd.write(str(auc))


if __name__ == "__main__":
    out = sys.argv[1]
    run_omp(out)
